"""
Tests for the ABCStatic class base
"""

# Standard
import abc

# Third Party
import pytest

# Local
from oper8.x.utils.abc_static import ABCStatic


class Base(ABCStatic):
    @classmethod
    def foo(cls):
        """A classmethod that is implemented in the base"""
        return True

    def bar(self):
        """A standard instance method on the base class"""
        return 42

    @classmethod
    @abc.abstractmethod
    def baz(cls, arg):
        """An abstractmethod that is also a classmethod"""

    @abc.abstractmethod
    def bat(self, arg):
        """An abstractmethod that is NOT a classmethod"""


## Tests #######################################################################


def test_correct_implementation():
    """Make sure that defining a class which follows the correct @classmethod
    definitions defines cleanly and can be used as expected
    """

    class Foo(Base):
        @classmethod
        def baz(cls, arg):
            return arg + 1

        def bat(self, arg):
            return arg + 2

    assert Foo.foo()
    assert Foo().bar() == 42
    assert Foo.baz(1) == 2
    assert Foo().bat(1) == 3


def test_staticmethod_implementation():
    """Test that an abstract classmethod can be implemented with a @staticmethod
    in the child
    """

    class Foo(Base):
        @staticmethod
        def baz(arg):
            return arg + 1

        def bat(self, arg):
            return arg + 2

    assert Foo.foo()
    assert Foo().bar() == 42
    assert Foo.baz(1) == 2
    assert Foo().bat(1) == 3


def test_classmethod_implemented_with_star_args():
    """Test that an abstract classmethod which is implemented with *args,
    **kwargs is ok
    """

    class Foo(Base):

        BASE = 10

        @classmethod
        def baz(cls, *args, **kwargs):
            return cls.BASE + 1

        def bat(self, arg):
            return arg + 2

    assert Foo.foo()
    assert Foo.baz(1) == 11


def test_incorrect_instancemethod_implementation():
    """Test that implementing an abstractmethod/classmethod as an instance
    method results in a declaration-time exception.
    """
    with pytest.raises(NotImplementedError):

        class Foo(Base):
            def __init__(self, x):
                self.x = x

            def baz(self, arg):
                return self.x + 1

            def bat(self, arg):
                return arg + 2


def test_abstractclassmethod_cannot_be_called():
    """Test that an abstractmethod/classmethod which is not implemented by a
    child class raises when called
    """

    class Foo(Base):
        pass

    with pytest.raises(NotImplementedError):
        Foo.baz(1)



"""
Tests of the common base factory
"""


# Third Party
import pytest

# First Party
import alog

# Local
from oper8 import component
from oper8.test_helpers.helpers import setup_session
from oper8.x.datastores.connection_base import DatastoreConnectionBase
from oper8.x.datastores.factory_base import DatastoreSingletonFactoryBase
from oper8.x.datastores.interfaces import Datastore
from oper8.x.utils import common, constants

## Helpers #####################################################################

log = alog.use_channel("TEST")

###################
## testDataStore ##
###################


class TestConnection(DatastoreConnectionBase):
    """Connection type for Test datastores"""

    __test__ = False

    def __init__(self, session, foo, bar):
        super().__init__(session)
        self.foo = foo
        self.bar = bar

    def to_dict(self):
        return {"foo": self.foo, "bar": self.bar}

    @classmethod
    def from_dict(cls, session, config_dict):
        return TestConnection(session, **config_dict)


class TestFactory(DatastoreSingletonFactoryBase):
    """Single factory for these tests
    NOTE: Since the global state is mutated at import time, we cannot do dynamic
        registration in the tests and must have a single class defined here
    """

    DATASTORE_TYPE = "testDataStore"
    CONNECTION_TYPE = TestConnection
    __test__ = False


class TestDatastoreBase(Datastore):
    __test__ = False

    def doit(self):
        return True


@component(name="test-datastore-one")
class TestDatastoreOne(TestDatastoreBase):

    TYPE_LABEL = "one"
    __test__ = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug("Constructing TestDatastoreOne")
        self.foo = False
        if self.config.do_it:
            self.add_resource(
                "secret-one",
                dict(
                    kind="Secret",
                    apiVersion="v1",
                    metadata=dict(name="secret-one"),
                    data=self.config.get("secret_data", {}),
                ),
            )
            self.foo = True

    def get_connection(self):
        return TestConnection(self.session, foo=self.foo, bar=1)


@component(name="test-datastore-two")
class TestDatastoreTwo(TestDatastoreBase):

    TYPE_LABEL = "two"
    __test__ = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug("Constructing TestDatastoreTwo")

    def get_connection(self):
        return TestConnection(self.session, foo=False, bar=2)


TestFactory.register_type(TestDatastoreOne)
TestFactory.register_type(TestDatastoreTwo)


#################
## widgetStore ##
#################


class WidgetConnection(DatastoreConnectionBase):
    """Widget store connection type"""

    __test__ = False

    def __init__(self, session, widgets):
        super().__init__(session)
        self.widgets = widgets

    def to_dict(self):
        return {"widgets": self.widgets}

    @classmethod
    def from_dict(cls, session, config_dict):
        return WidgetConnection(session, **config_dict)


class WidgetFactory(DatastoreSingletonFactoryBase):
    """Second datastore factory type"""

    DATASTORE_TYPE = "widgetStore"
    CONNECTION_TYPE = WidgetConnection


class WidgetStoreBase(Datastore):
    pass


@component(name="widget-store-a")
class WidgetStoreA(WidgetStoreBase):

    TYPE_LABEL = "A"
    __test__ = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        log.debug("Constructing TestDatastoreOne")
        config = self.config
        self.widgets = 0
        if self.config.do_it:
            self.widgets += 1
            self.add_resource(
                "secret-one",
                dict(
                    kind="Secret",
                    apiVersion="v1",
                    metadata=dict(name="secret-one"),
                    data=self.config.get("secret_data", {}),
                ),
            )

    def get_connection(self):
        return WidgetConnection(self.session, self.widgets)


WidgetFactory.register_type(WidgetStoreA)


## Tests #######################################################################

###################
## get_component ##
###################


def test_get_component_construct_by_type():
    """Test that a basic construction works"""
    instance1_name = "test_instance_one"
    instance1_config = {
        "type": "one",
        "key": "val",
    }
    instance2_name = "test_instance_two"
    instance2_config = {
        "type": "two",
        "key": "val",
    }

    _config = {
        TestFactory.DATASTORE_TYPE: {
            instance1_name: instance1_config,
            instance2_name: instance2_config,
        }
    }
    session = setup_session(app_config=_config)

    # Instance 1 is of type "one"
    instance1 = TestFactory.get_component(session, instance1_name)
    assert isinstance(instance1, TestDatastoreOne)
    assert instance1.config == instance1_config

    # Instance 2 is of type "two"
    instance2 = TestFactory.get_component(session, instance2_name)
    assert isinstance(instance2, TestDatastoreTwo)
    assert instance2.config == instance2_config


def test_get_component_base_class_function():
    """Test that a constructed instance can use a base class's function"""
    instance_name = "test_instance_one"
    instance_config = {
        "type": "one",
        "key": "val",
    }
    _config = {
        TestFactory.DATASTORE_TYPE: {
            instance_name: instance_config,
        }
    }
    session = setup_session(app_config=_config)

    instance = TestFactory.get_component(session, instance_name)
    assert instance.doit()


def test_get_component_derived_class_config_use():
    """Test that a derived class can use the session config to take conditional
    action
    """
    instance_name = "test_instance_one"
    secret_data = {"secret": common.b64_secret("value")}

    def app_config_overrides(doit):
        return {
            TestFactory.DATASTORE_TYPE: {
                instance_name: {
                    "type": "one",
                    "do_it": doit,
                    "secret_data": secret_data,
                },
            }
        }

    # Validate that with "do_it" True a secret is created
    with alog.ContextLog(log.debug, "Enabled"):
        _config = app_config_overrides(True)
        session = setup_session(app_config=_config)
        instance = TestFactory.get_component(session, instance_name)
        resources = instance.to_config(session)
        assert len(resources) == 1
        assert resources[0].data == secret_data

    # Validate that with "do_it" False the secret is not created
    with alog.ContextLog(log.debug, "Disabled"):
        _config = app_config_overrides(False)
        session = setup_session(app_config=_config)
        instance = TestFactory.get_component(session, instance_name)
        resources = instance.to_config(session)
        assert len(resources) == 0


def test_get_component_per_deploy_singleton():
    """Test that the same exact instance is returned for multiple calls within
    a single deploy, but across deploys the instance is recreated
    """
    instance_name = "test_instance_one"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    instance_a = TestFactory.get_component(session, instance_name)
    instance_b = TestFactory.get_component(session, instance_name)
    assert id(instance_a) == id(instance_b)

    session = setup_session(app_config=app_config_overrides)
    instance_c = TestFactory.get_component(session, instance_name)
    assert id(instance_a) != id(instance_c)


def test_get_component_multi_datastore_type():
    """Test that the factory can handle multiple datastore types independently"""
    test_instance_name = "test_instance_one"
    widget_component_name = "widget_component_a"
    widget_bad_instance_name = "widget_component_one"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {test_instance_name: {"type": "one"}},
        WidgetFactory.DATASTORE_TYPE: {
            widget_component_name: {"type": "A"},
            widget_bad_instance_name: {"type": "one"},
        },
    }
    session = setup_session(app_config=app_config_overrides)
    test_one = TestFactory.get_component(session, test_instance_name)
    assert test_one is not None
    widget_a = WidgetFactory.get_component(session, widget_component_name)
    assert widget_a is not None
    with pytest.raises(AssertionError):
        WidgetFactory.get_component(session, widget_bad_instance_name)


def test_get_component_unnamed():
    """Test that the factory can fetch an un-named instance"""
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {"type": "one"},
    }
    session = setup_session(app_config=app_config_overrides)
    test_one = TestFactory.get_component(session)
    assert test_one is not None
    assert test_one.__class__.name == TestDatastoreOne.name


def test_get_component_named_class_override():
    """Test that the factory can create an instance with an overloaded class
    name that is scoped by the instance name
    """
    instance_name = "foo"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    test_one = TestFactory.get_component(session, instance_name)
    assert test_one is not None
    assert test_one.__class__.name == f"{TestDatastoreOne.name}-{instance_name}"


def test_get_component_disabled():
    """Test that a constructed instance which is disabled is noted as such in
    the session
    """
    instance_name = "foo"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    test_one = TestFactory.get_component(session, instance_name, disabled=True)
    assert test_one is not None
    assert test_one.__class__.name == f"{TestDatastoreOne.name}-{instance_name}"
    enabled_components = session.get_components(disabled=False)
    assert not enabled_components
    disabled_components = session.get_components(disabled=True)
    assert disabled_components == [test_one]


def test_get_component_connection_in_cr():
    """Test that when asked for a component, but connection details are provided
    in the CR, no component is constructed
    """
    instance_name = "foo"
    input_config = {"foo": True, "bar": 2}
    deploy_config_overrides = {
        "datastores": {
            TestFactory.DATASTORE_TYPE: {
                instance_name: {constants.SPEC_DATASTORE_CONNECTION: input_config}
            }
        }
    }
    session = setup_session(deploy_config=deploy_config_overrides)
    test_one = TestFactory.get_component(session, instance_name, disabled=True)
    assert test_one is None


def test_get_component_config_overrides():
    """Test that overrides given in config_overrides take precidence over
    session.config
    """
    instance_name = "foo"
    app_config_overrides = {TestFactory.DATASTORE_TYPE: {"type": "one"}}
    session = setup_session(app_config=app_config_overrides)
    instance = TestFactory.get_component(
        session,
        config_overrides={
            "type": "two",
        },
    )
    assert isinstance(instance, TestDatastoreTwo)


def test_get_component_named_config_overrides():
    """Test that overrides given in config_overrides with instance name nesting
    take precidence over session.config
    """
    instance_name = "foo"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    instance = TestFactory.get_component(
        session,
        instance_name,
        config_overrides={
            instance_name: {"type": "two"},
        },
    )
    assert isinstance(instance, TestDatastoreTwo)


####################
## get_connection ##
####################


def test_get_connection_from_component():
    """Test that fetching a connection after constructing the component yields
    a connection which matches the one returned by the component
    """
    instance_name = "test_instance_one"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    instance = TestFactory.get_component(session, instance_name)
    connection = TestFactory.get_connection(session, instance_name)
    assert connection is not None
    assert connection.to_dict() == instance.get_connection().to_dict()


def test_get_connection_from_config():
    """Test that fetching a connection without instantiating a component pulls
    from the CR config
    """
    instance_name = "test_instance_one"
    input_config = {"foo": True, "bar": 2}
    deploy_config_overrides = {
        "datastores": {
            TestFactory.DATASTORE_TYPE: {
                instance_name: {constants.SPEC_DATASTORE_CONNECTION: input_config}
            }
        }
    }
    session = setup_session(deploy_config=deploy_config_overrides)
    connection = TestFactory.get_connection(session, instance_name)
    assert connection is not None
    assert connection.to_dict() == input_config


def test_get_connection_per_deploy_singleton():
    """Test that fetching a config multiple times results in the exact same
    instance within a single deploy, but different instances between deploys
    """
    instance_name = "test_instance_one"
    input_config_one = {"foo": True, "bar": 1}
    input_config_two = {"foo": False, "bar": 2}
    session = setup_session(
        deploy_config={
            "datastores": {
                TestFactory.DATASTORE_TYPE: {
                    instance_name: {
                        constants.SPEC_DATASTORE_CONNECTION: input_config_one
                    }
                }
            }
        },
    )
    connection_one = TestFactory.get_connection(session, instance_name)
    assert connection_one is not None
    assert connection_one.to_dict() == input_config_one
    connection_two = TestFactory.get_connection(session, instance_name)
    assert id(connection_one) == id(connection_two)

    session = setup_session(
        deploy_config={
            "datastores": {
                TestFactory.DATASTORE_TYPE: {
                    instance_name: {
                        constants.SPEC_DATASTORE_CONNECTION: input_config_two
                    }
                }
            }
        },
    )
    connection_three = TestFactory.get_connection(session, instance_name)
    assert connection_one is not None
    assert connection_three.to_dict() != connection_one.to_dict()


def test_get_connection_unnamed_from_config():
    """Test that fetching an unnamed connection from config works"""
    input_config = {"foo": True, "bar": 2}
    deploy_config_overrides = {
        "datastores": {
            TestFactory.DATASTORE_TYPE: {
                constants.SPEC_DATASTORE_CONNECTION: input_config
            }
        }
    }
    session = setup_session(deploy_config=deploy_config_overrides)
    connection = TestFactory.get_connection(session)
    assert connection is not None
    assert connection.to_dict() == input_config


def test_get_connection_unnamed_from_instance():
    """Test that fetching an unnamed connection from an unnamed instance works"""
    app_config_overrides = {TestFactory.DATASTORE_TYPE: {"type": "one"}}
    session = setup_session(app_config=app_config_overrides)
    instance = TestFactory.get_component(session)
    connection = TestFactory.get_connection(session)
    assert connection is not None
    assert connection.to_dict() == instance.get_connection().to_dict()


def test_get_connection_missing_config():
    """Test that trying to fetch a connection from config asserts if the config
    is missing and no instance was populated from the CR
    """
    session = setup_session()
    with pytest.raises(AssertionError):
        TestFactory.get_connection(session)
    with pytest.raises(AssertionError):
        TestFactory.get_connection(session, "foobar")


def test_get_connection_no_config_or_instance():
    """Test that attempting to fetch a connection where there is no config or
    instance available raises an assertion
    """
    instance_name = "test_instance_one"
    app_config_overrides = {
        TestFactory.DATASTORE_TYPE: {instance_name: {"type": "one"}}
    }
    session = setup_session(app_config=app_config_overrides)
    with pytest.raises(AssertionError):
        TestFactory.get_connection(session, instance_name)


def test_reregister_ok():
    """Make sure that a type can be re-registered without raising. This is
    needed when registration is done in a derived library that uses the PWM and
    therefore re-imports the derived implementation.
    """
    TestFactory.register_type(TestDatastoreOne)



"""
Filters are used to limit the amount of events being reconciled by a watch manager
This is based off of the kubernetes controller runtime's "predicates":
https://pkg.go.dev/sigs.k8s.io/controller-runtime@v0.15.0/pkg/predicate#Funcs
The default set of filters is derived from operator-sdk's ansible predicates
https://github.com/operator-framework/operator-sdk/blob/50c6ac03746ff4edf582feb9a71d2a7ea6ae6c40/internal/ansible/controller/controller.go#L105
"""

# Standard
from abc import ABC, abstractmethod
from collections import deque
from typing import Optional

# First Party
import alog

# Local
from ....deploy_manager import KubeEventType
from ....managed_object import ManagedObject
from ....reconcile import ReconcileManager
from ....status import READY_CONDITION, get_condition
from ....utils import abstractclassproperty
from ..utils import (
    RESERVED_PLATFORM_ANNOTATIONS,
    RESOURCE_VERSION_KEEP_COUNT,
    obj_to_hash,
)

log = alog.use_channel("PWMFLT")


## Default Types


class Filter(ABC):
    """Generic Filter Interface for subclassing. Every subclass should implement a
    `test` function which returns true when a resource should be reconciled. Subclasses
    can optionally implement a `update` method if the filter requires storing some stateful
    information like ResourceVersion or Metadata.

    NOTE: A unique Filter instance is created for each resource
    """

    def __init__(self, resource: ManagedObject):  # noqa: B027
        """Initializer can be used to detect configuration or create instance
        variables. Even though a resource is provided it should not set state until
        update is called

        Args:
            resource: ManagedObject
                This resource can be used by subclass to gather generic information.

        """

    ## Abstract Interface ######################################################
    #
    # These functions must be implemented by child classes
    ##

    @abstractmethod
    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Test whether the resource&event passes the filter. Returns true if
        the filter should be reconciled and return false if it should not be. A filter
        can optionally return None to ignore an event

        Args:
            resource: ManagedObject
                The current resource being checked
            event: KubeEventType
                The event type that triggered this filter

        Returns:
            result: Optional[bool]
                The result of the test.

        """

    ## Base Class Interface ####################################################
    #
    # These methods MAY be implemented by children, but contain default
    # implementations that are appropriate for simple cases.
    #
    ##

    def update(self, resource: ManagedObject):  # noqa: B027
        """Update the instances current state.

        Args:
            resource: ManagedObject
               The current state of the resource
        """

    def update_and_test(self, resource: ManagedObject, event: KubeEventType) -> bool:
        """First test a resource/event against a filter then update the current state

        Args:
            resource: ManagedObject
                The resource being filtered
            event: KubeEventType
                The event to be filtered

        Returns:
            test_result: bool
                The test result
        """
        result = self.test(resource, event)
        if result is not None and not result:
            log.debug3(
                "Failed filter: %s with return val %s",
                self,
                result,
                extra={"resource": resource},
            )
        self.update(resource)
        return result


## Generic Resource filters


class CreationDeletionFilter(Filter):
    """Filter to ensure reconciliation on creation and deletion events"""

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Return true if event is ADDED or DELETED"""

        # Ignore non Added/Deleted Events
        if event not in [KubeEventType.ADDED, KubeEventType.DELETED]:
            return

        return True


class GenerationFilter(Filter):
    """Filter for reconciling on generation changes for resources that support it"""

    def __init__(self, resource: ManagedObject):
        """Set generation instance variable"""
        super().__init__(resource)
        self.generation = None

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Return true if resource generation is different than before"""
        # Only update&test resources with a generation
        if not self.generation:
            return

        # Only test on resource updates
        if event in [KubeEventType.ADDED, KubeEventType.DELETED]:
            return

        # Test if new generation is different
        return self.generation != resource.metadata.get("generation")

    def update(self, resource: ManagedObject):
        """Update the currently observed generation"""
        self.generation = resource.metadata.get("generation")


class NoGenerationFilter(Filter):
    """Filter for reconciling changes to spec on resources that don't support
    the generation field like pods. It does this by hashing the objects excluding
    status and metadata"""

    def __init__(self, resource: ManagedObject):
        """Check if resource supports generation and initialize the hash dict"""
        self.supports_generation = resource.metadata.get("generation") is not None
        self.resource_hashes = {}
        super().__init__(resource)

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Return True if a resources current hash differs from the current"""
        # Don't test resources that support generation or if we don't have hashes yet
        if self.supports_generation or not self.resource_hashes:
            return

        # Only test on resource updates
        if event in [KubeEventType.ADDED, KubeEventType.DELETED]:
            return

        # Check each stored resource hash to see if its
        # changed
        for key, obj_has in self.resource_hashes.items():
            if obj_has != obj_to_hash(resource.get(key)):
                log.debug2("Detected change in %s", key)
                return True

        return False

    def update(self, resource: ManagedObject):
        """Update the observed spec hashes"""
        if self.supports_generation:
            return

        # Get the default hashes for all object keys except metadata
        # and status
        for key, obj in resource.definition.items():
            if key in ["metadata", "status", "kind", "apiVersion"]:
                continue

            self.resource_hashes[key] = obj_to_hash(obj)


class ResourceVersionFilter(Filter):
    """Filter for duplicate resource versions which happens when restarting a
    watch connection"""

    def __init__(self, resource: ManagedObject):
        """Initialize the resource version list"""
        # Use a dequeue instead of a list/set to set a bound on the number
        # of tracked versions
        self.resource_versions = deque([], maxlen=RESOURCE_VERSION_KEEP_COUNT)
        super().__init__(resource)

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Test if the resource's resourceVersion has been seen before"""

        # Don't skip add events as the kubernetes watch can duplicate events
        if event == KubeEventType.DELETED:
            return

        return resource.resource_version not in self.resource_versions

    def update(self, resource: ManagedObject):
        """Add the resources ResourceVersion to the list"""
        self.resource_versions.append(resource.resource_version)


### Annotation Filters


class AnnotationFilter(Filter):
    """Filter resources to reconcile on annotation changes"""

    def __init__(self, resource: ManagedObject):
        """Initialize the annotation hash variable"""
        self.annotations = None
        super().__init__(resource)

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Test if a resource's annotation has changed"""
        # Ignore Added and deleted events
        if event in [KubeEventType.ADDED, KubeEventType.DELETED]:
            return

        return self.annotations != self.get_annotation_hash(resource)

    def update(self, resource: ManagedObject):
        """Update the currently stored annotation"""
        self.annotations = self.get_annotation_hash(resource)

    def get_annotation_hash(self, resource: ManagedObject) -> str:
        """Helper function to get the annotation hash"""
        return obj_to_hash(resource.metadata.get("annotations", {}))


class UserAnnotationFilter(AnnotationFilter):
    """Filter resources to reconcile on user annotation changes. This excludes
    kubernetes and openshift annotations
    """

    def get_annotation_hash(self, resource: ManagedObject) -> str:
        """Overriden function to exclude common platform annotations from
        the annotation hash"""
        output_annotations = {}
        for key, value in resource.metadata.get("annotations", {}).items():
            if self.contains_platform_key(key):
                continue

            output_annotations[key] = value

        return obj_to_hash(output_annotations)

    def contains_platform_key(self, key: str) -> bool:
        """Helper to check if the key contains one of the
        platform annotations"""
        return any(
            reserved_key in key for reserved_key in RESERVED_PLATFORM_ANNOTATIONS
        )


### Oper8 Filters


class PauseFilter(Filter):
    """This filter skips resources that have the oper8 pause annotation"""

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Test if a resource has the pause annotation"""
        return not ReconcileManager._is_paused(  # pylint: disable=protected-access
            resource
        )


class SubsystemStatusFilter(Filter):
    """Reconcile oper8 controllers when their oper8 status changes

    EXPERIMENTAL: This has passed basic validation but has not been rigorously tested
     in the field
    """

    def __init__(self, resource: ManagedObject):
        """Initialize the currently observed ready condition"""
        self.ready_condition = None
        super().__init__(resource)

    def test(  # pylint: disable=inconsistent-return-statements
        self,
        resource: ManagedObject,
        event: KubeEventType,
    ) -> Optional[bool]:
        """Test if a resources subsystem condition has changed"""
        if event in [KubeEventType.ADDED, KubeEventType.DELETED]:
            return

        return self.ready_condition != get_condition(
            READY_CONDITION, resource.get("status", {})
        ).get("reason")

    def update(self, resource: ManagedObject):
        """Update the currently observed ready condition"""
        self.ready_condition = get_condition(
            READY_CONDITION, resource.get("status", {})
        ).get("reason")


### Dependent Filters


class DependentWatchFilter(Filter):
    """Don't reconcile creation events as we can assume the owner created
    them"""

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Return False if event is ADDED"""
        return event != KubeEventType.ADDED


### Utility Filters


class LabelFilter(Filter):
    """Filter for resources that match a set of labels"""

    @abstractclassproperty
    def labels(self) -> dict:
        """Subclasses must implement a labels class attribute"""

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Return true is a resource matches the requested labels"""
        resource_labels = resource.get("metadata", {}).get("labels")
        # Check to make sure every requested label matches
        return all(
            resource_labels.get(label) == value for label, value in self.labels.items()
        )


class DisableFilter(Filter):
    """Filter to disable all reconciles"""

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Always return False"""
        return False


class EnableFilter(Filter):
    """Filter to run all reconciles"""

    def test(self, resource: ManagedObject, event: KubeEventType) -> Optional[bool]:
        """Always return True"""
        return True



"""Base classes for leader election implementations"""

# Standard
from typing import Optional
import abc
import threading

# First Party
import alog

# Local
from .... import config
from ....deploy_manager import DeployManagerBase
from ....exceptions import ConfigError
from ....managed_object import ManagedObject
from ..utils import ABCSingletonMeta, parse_time_delta

log = alog.use_channel("LDRELC")


class LeadershipManagerBase(abc.ABC):
    """
    Base class for leader election. Leadership election in the PWM
    is split into two types: global and resource locks. Global locks
    are required to run any reconciliation while resource locks are
    required to reconcile a specific resources. Most child classes
    implement one of these locks.
    """

    def __init__(self, deploy_manager: DeployManagerBase = None):
        """
        Initialize Class

        Args:
            deploy_manager:  DeployManagerBase
                DeployManager used in lock acquisition
        """
        self.deploy_manager = deploy_manager

    ## Lock Interface ####################################################
    @abc.abstractmethod
    def acquire(self, force: bool = False) -> bool:
        """
        Acquire or renew global lock

        Args:
            force:  bool
                Whether to force acquire the lock irregardless of status. Used
                on shutdown

        Returns:
            success:  bool
                True on successful acquisition
        """

    @abc.abstractmethod
    def acquire_resource(self, resource: ManagedObject) -> bool:
        """
        Acquire or renew lock on specific resource

        Args:
            resource:  ManagedObject
                Resource to acquire lock for
        Returns:
            success:  bool
                True on successful acquisition
        """

    @abc.abstractmethod
    def release(self):
        """
        Release global lock
        """

    @abc.abstractmethod
    def release_resource(self, resource: ManagedObject):
        """
        Release lock on specific resource

        Args:
            resource:  ManagedObject
                Resource to release lock for
        """

    @abc.abstractmethod
    def is_leader(self, resource: Optional[ManagedObject] = None):
        """
        Determines if current instance is leader

        Args:
            resource:  Optional[ManagedObject]
                If provided the resource to determine if current instance
                is leader for. If no resource if provided then the global
                lock is checked
        Returns:
            leader:  bool
                True if instance is leader
        """


class ThreadedLeaderManagerBase(LeadershipManagerBase, metaclass=ABCSingletonMeta):
    """
    Base class for threaded leadership election. This base class aids in the
    creation of leadership election classes that require constantly checking
    or updating a resource. Child classes only need to implement renew_or_acquire,
    and it will automatically be looped while lock acquisition is needed
    """

    def __init__(self, deploy_manager: DeployManagerBase):
        """
        Initialize class with events to track leadership and shutdown and
        a lock to ensure renew_or_acquire is only ran once.

        Args:
            deploy_manager: DeployManagerBase
                DeployManager for leader election
        """
        super().__init__(deploy_manager)

        # Events to track status
        self.leader = threading.Event()
        self.shutdown = threading.Event()

        # Lock to ensure multiple acquires aren't running at the same time
        self.run_lock = threading.Lock()

        # Object to track Leadership thread
        self.leadership_thread = threading.Thread(
            name="leadership_thread", target=self.run, daemon=True
        )

        # Calculate threaded poll time:
        poll_time_delta = parse_time_delta(config.python_watch_manager.lock.poll_time)
        if not poll_time_delta:
            log.error(
                "Invalid 'python_watch_manager.lock.poll_time' value: '%s'",
                config.python_watch_manager.lock.poll_time,
            )
            raise ConfigError(
                "Invalid 'python_watch_manager.lock.poll_time' value: "
                f"'{config.python_watch_manager.lock.poll_time}'"
            )
        self.poll_time = poll_time_delta.seconds

    ## Public Interface ####################################################

    def renew_or_acquire(self):
        """
        Renew or acquire leadership lock
        """
        raise NotImplementedError

    def acquire_lock(self):
        """
        Helper function for child classes to acquire leadership lock
        """
        if not self.leader.is_set():
            log.debug2("Acquiring leadership lock")
        # Always set the lock during acquire_lock to avoid concurrency issues
        self.leader.set()

    def release_lock(self):
        """
        Helper function for child classes to release lock
        """
        if self.leader.is_set():
            log.debug2("Releasing leadership lock")
        self.leader.clear()

    ## Lock Interface ####################################################
    def acquire(self, force: bool = False):
        """
        Start/Restart leadership thread or run renew_or_acquire

        Args:
            force:  bool=False
                Whether to force acquire the lock

        Returns:
            success:  bool
                True on successful acquisition
        """
        if force:
            self.leader.set()
            return True

        # ident is set when thread has started
        if not self.leadership_thread.is_alive():
            # Recreate leadership thread if its already exited
            if self.leadership_thread.ident:
                self.leadership_thread = threading.Thread(
                    name="leadership_thread", target=self.run, daemon=True
                )
            log.info(
                "Starting %s: %s", self.__class__.__name__, self.leadership_thread.name
            )
            self.leadership_thread.start()
        else:
            self.run_renew_or_acquire()

        return self.leader.wait()

    def acquire_resource(self, resource: ManagedObject) -> bool:
        """
        Lock in background so acquire_resource just waits for value

        Args:
            resource:  ManagedObject
                Resource that is being locked

        Returns:
            success:  bool
                True on successful acquisition else False
        """
        return self.leader.wait()

    def release(self):
        """
        Release lock and shutdown leader election thread. This thread
        first shuts down the background thread before clearing the lock
        """
        self.shutdown.set()
        self.leadership_thread.join()
        self.leader.clear()

    def release_resource(self, resource: ManagedObject):
        """
        Release resource is not implemented in Threaded classes
        """

    def is_leader(self, resource: Optional[ManagedObject] = None) -> bool:
        """
        Return if leader event has been acquired

        Returns:
            leader: bool
                If instance is current leader
        """
        return self.leader.is_set()

    ## Implementation Details ####################################################

    def run(self):
        """
        Loop to continuously run renew or acquire every so often
        """
        while True:
            if self.shutdown.is_set():
                log.debug("Shutting down %s Thread", self.__class__.__name__)
                return

            self.run_renew_or_acquire()
            self.shutdown.wait(self.poll_time)

    def run_renew_or_acquire(self):
        """
        Run renew_or_acquire safely and with threaded lock
        """
        log.debug2("Running renew or acquire for %s lock", self.__class__.__name__)
        with self.run_lock:
            try:
                self.renew_or_acquire()
            except Exception as err:
                log.warning(
                    "Error detected while acquiring leadership lock", exc_info=True
                )
                raise RuntimeError("Error detected when acquiring lock") from err



"""
Package exports
"""

# Local
from . import config, reconcile, status, watch_manager
from .component import Component
from .controller import Controller
from .dag import Graph, ResourceNode
from .decorator import component, controller
from .deploy_manager import DeployManagerBase
from .exceptions import (
    assert_cluster,
    assert_config,
    assert_precondition,
    assert_verified,
)
from .reconcile import ReconcileManager, ReconciliationResult
from .session import Session
from .temporary_patch.temporary_patch_controller import TemporaryPatchController
from .verify_resources import verify_resource



"""
Tests for the update_owner_references functionality
"""

# Standard
import copy

# Third Party
import pytest

# First Party
import alog

# Local
from oper8.deploy_manager.owner_references import (
    _make_owner_reference,
    update_owner_references,
)
from oper8.exceptions import ClusterError
from oper8.test_helpers.helpers import (
    SOME_OTHER_NAMESPACE,
    TEST_NAMESPACE,
    MockDeployManager,
)

## Helpers #####################################################################

log = alog.use_channel("TEST")

SAMPLE_OWNER = {
    "kind": "Owner",
    "apiVersion": "foo.bar.com/v1",
    "metadata": {
        "name": "owner",
        "namespace": TEST_NAMESPACE,
        "uid": "12345",
    },
}


def sample_object(namespace=TEST_NAMESPACE):
    return {
        "kind": "Child",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {
            "name": "child",
            "namespace": namespace,
            "uid": "54321",
        },
    }


## Happy Path ##################################################################


def test_add_new_owner_ref():
    """Test that adding a ref to an object with none present adds as expected"""
    dm = MockDeployManager()
    obj = sample_object()
    update_owner_references(dm, SAMPLE_OWNER, obj)
    assert "ownerReferences" in obj["metadata"]
    assert obj["metadata"]["ownerReferences"] == [_make_owner_reference(SAMPLE_OWNER)]


def test_do_not_add_new_owner_ref_if_namespace_mismatch():
    """Test that a ref to an object with none present is only added if owner and object are in same namespace"""
    dm = MockDeployManager()
    obj = sample_object(namespace=SOME_OTHER_NAMESPACE)
    update_owner_references(dm, SAMPLE_OWNER, obj)
    assert "ownerReferences" in obj["metadata"]
    assert obj["metadata"]["ownerReferences"] == []


def test_no_duplicate():
    """Test that an object with an existing ref for the owner does not
    duplicate the existing ref
    """
    dm = MockDeployManager()
    obj = sample_object()
    obj["metadata"]["ownerReferences"] = [_make_owner_reference(SAMPLE_OWNER)]
    update_owner_references(dm, SAMPLE_OWNER, obj)
    assert "ownerReferences" in obj["metadata"]
    assert obj["metadata"]["ownerReferences"] == [_make_owner_reference(SAMPLE_OWNER)]


def test_external_preserved():
    """Test that an object with an existing ref for a different owner adds the
    new reference without removing the old one
    """
    external_owner = {
        "kind": "Owner",
        "apiVersion": "foo.bar.com/v1",
        "metadata": {
            "name": "other-owner",
            "namespace": "test",
            "uid": "67890",
        },
    }
    external_ref = _make_owner_reference(external_owner)
    obj = sample_object()
    cluster_content_obj = copy.deepcopy(obj)
    cluster_content_obj["metadata"]["ownerReferences"] = [external_ref]
    dm = MockDeployManager(resources=[cluster_content_obj])
    update_owner_references(dm, SAMPLE_OWNER, obj)
    assert "ownerReferences" in obj["metadata"]
    assert obj["metadata"]["ownerReferences"] == [
        external_ref,
        _make_owner_reference(SAMPLE_OWNER),
    ]


## Happy Path ##################################################################


def test_owner_missing_keys():
    """Test that when the owner is missing required keys, an assertion is hit"""
    dm = MockDeployManager()
    for key in ["kind", "apiVersion", "metadata.name", "metadata.namespace"]:
        log.debug("Trying without [%s]", key)
        bad_owner = copy.deepcopy(SAMPLE_OWNER)
        parts = key.split(".")
        dct = bad_owner
        for part in parts[:-1]:
            dct = dct[part]
        del dct[parts[-1]]
        with pytest.raises(AssertionError):
            update_owner_references(dm, bad_owner, sample_object())


def test_child_missing_keys():
    """Test that when the child is missing required keys, an assertion is hit"""
    dm = MockDeployManager()
    for key in ["kind", "apiVersion", "metadata.name", "metadata.namespace"]:
        log.debug("Trying without [%s]", key)
        bad_child = sample_object()
        parts = key.split(".")
        dct = bad_child
        for part in parts[:-1]:
            dct = dct[part]
        del dct[parts[-1]]
        with pytest.raises(AssertionError):
            update_owner_references(dm, SAMPLE_OWNER, bad_child)


def test_lookup_cluster_error():
    """Test that when the deploy manager fails to look up the object, a
    ClusterError is raised
    """
    dm = MockDeployManager(get_state_fail=True)
    obj = sample_object()
    with pytest.raises(ClusterError):
        update_owner_references(dm, SAMPLE_OWNER, obj)


## _make_owner_reference #######################################################


def test_make_owner_reference_required_keys():
    """Make sure the shape of the owner reference looks right"""
    ref = _make_owner_reference(SAMPLE_OWNER)
    for key, source in {
        "apiVersion": "apiVersion",
        "kind": "kind",
        "name": "metadata.name",
        "uid": "metadata.uid",
    }.items():
        parts = source.split(".")
        dct = SAMPLE_OWNER
        for part in parts[:-1]:
            dct = dct[part]
        exp_val = dct[parts[-1]]
        assert key in ref
        assert ref[key] == exp_val


def test_make_owner_block_owner_deletion():
    """Make sure that blockOwnerDeletion is set"""
    ref = _make_owner_reference(SAMPLE_OWNER)
    assert "blockOwnerDeletion" in ref
    assert ref["blockOwnerDeletion"]


def test_make_owner_not_controller():
    """Make sure that controller is not set"""
    ref = _make_owner_reference(SAMPLE_OWNER)
    assert "controller" not in ref



"""Implementation of the DryRun LeaderElection"""
# Standard
from typing import Optional

# Local
from ....managed_object import ManagedObject
from .base import LeadershipManagerBase


class DryRunLeadershipManager(LeadershipManagerBase):
    """DryRunLeaderElection class implements an empty leadership
    election manager which always acts as a leader. This is useful
    for dryrun or running without leadership election"""

    def acquire(self, force: bool = False):
        """
        Return true as dryrun is always leader
        """
        return True

    def acquire_resource(self, resource: ManagedObject):
        """
        Return true as dryrun is always leader
        """
        return True

    def release(self):
        """
        NoOp in DryRun as lock is not real
        """

    def release_resource(self, resource: ManagedObject):
        """
        NoOp in DryRun as lock is not real
        """

    def is_leader(self, resource: Optional[ManagedObject] = None):
        """
        DryRunLeadershipManager is always leader
        """
        return True



"""
Tests for the common filter functions
"""
# Third Party
import pytest

# First Party
import alog

# Local
from oper8.exceptions import ConfigError
from oper8.watch_manager.python_watch_manager.filters.common import import_filter
from oper8.watch_manager.python_watch_manager.filters.filters import EnableFilter

## Helpers #####################################################################


def test_import_filter():
    assert (
        import_filter(
            "oper8.watch_manager.python_watch_manager.filters.filters.EnableFilter"
        )
        == EnableFilter
    )


def test_import_filter_fail():
    with pytest.raises(ConfigError):
        import_filter(
            "oper8.watch_manager.python_watch_manager.filters.filters.DoesNotExist"
        )



"""
Tests for setting up a VCS repo
"""
# Standard
from dataclasses import dataclass, field
from typing import Dict, List, Optional
import os
import tempfile
import time

# Third Party
import pygit2
import pytest

# Local
from oper8.setup_vcs import setup_vcs

## Helpers #####################################################################

TEST = pygit2.Signature("Test", "test@test.it")


@dataclass
class TreeNode:
    files: Dict[str, str]
    message: str = "stub msg"
    children: List["TreeNode"] = field(default_factory=list)
    branch_name: Optional[str] = None
    tag_name: Optional[str] = None


def make_commits(
    repo: pygit2.Repository, node: TreeNode, parent: Optional[pygit2.Reference] = None
):
    # Create the files
    repo_root_path = os.path.realpath(os.path.join(repo.path, ".."))
    for fname, content in node.files.items():
        fdir, file_name = os.path.split(fname)
        base_dir = repo_root_path
        if fdir:
            base_dir = os.path.join(repo_root_path, fdir)
            os.makedirs(base_dir, exist_ok=True)
        with open(os.path.join(base_dir, file_name), "w") as handle:
            handle.write(content)
        repo.index.read()
        repo.index.add(fname)
        repo.index.write()

    # Make the commit
    tree = repo.index.write_tree()
    parents = [parent.target] if parent is not None else []
    commit_oid = repo.create_commit("HEAD", TEST, TEST, node.message, tree, parents)
    commit_ref = repo.head

    # If a branch name is given, make the branch, but don't check it out
    if node.branch_name:
        repo.create_branch(node.branch_name, repo.get(commit_oid))

    # If a tag name is given, make the tag
    if node.tag_name:
        repo.create_tag(
            node.tag_name,
            commit_oid,
            1,
            TEST,
            "some tag",
        )

    # Recurse on all children
    for child in node.children:
        make_commits(repo, child, commit_ref)

    # Reset the main branch to the parent and then check it out
    if parent:
        main_branch = repo.branches["main"]
        main_branch.set_target(parent.target)
        repo.checkout(main_branch)
        try:
            repo.stash(TEST, include_untracked=True)
            repo.stash_drop()
        except KeyError:
            pass


@pytest.fixture
def workdir():
    with tempfile.TemporaryDirectory() as workdir:
        prev_dir = os.getcwd()
        os.chdir(workdir)
        try:
            yield workdir
        finally:
            os.chdir(prev_dir)


def temp_git_repo(parent_dir: str, repo_root: TreeNode):
    repo_path = os.path.join(parent_dir, "repo")
    repo = pygit2.init_repository(repo_path, initial_head="main")
    make_commits(repo, repo_root)
    return repo_path


def validate_vcs_dir(
    source_repo_path: str, dest_repo_path: str, expected_branches: List[str]
):
    dest_repo = pygit2.Repository(dest_repo_path)
    root_commit = dest_repo.get(dest_repo.head.target)
    assert not root_commit.parents

    # Get all branches and make sure the list matches
    all_branches = set(dest_repo.branches)
    assert all_branches == set(expected_branches)

    # Add the source as a remote so we can compare files
    remote = dest_repo.remotes.create("validate", source_repo_path)
    progress = remote.fetch()
    while progress.received_objects < progress.total_objects:
        time.sleep(0.001)

    # For each branch validate that it's exactly one hop from the root and that
    # the content matches the source
    for branch_name in expected_branches:
        branch = dest_repo.branches.get(branch_name)
        branch_commit = dest_repo.get(branch.target)

        # One commit removed from root
        assert branch_commit.parents == [root_commit]

        # Matches remote
        source_ref = dest_repo.references.get(f"refs/tags/{branch_name}")
        if not source_ref:
            source_ref = dest_repo.references.get(
                f"refs/remotes/validate/{branch_name}"
            )
        assert source_ref
        source_commit = dest_repo.get(source_ref.target)
        if not isinstance(source_commit, pygit2.Commit):
            source_commit = dest_repo.get(source_commit.target)
        dest_repo.checkout(source_ref)
        diff = branch_commit.tree.diff_to_workdir()
        assert not list(diff.deltas)


TEST_REPO_TREE = TreeNode(
    {"foo.txt": "Hello!"},
    "Root commit",
    [
        TreeNode({"foo.txt": "Hello World!"}, "some branch", branch_name="branch/one"),
        TreeNode(
            {"bar.txt": "Hey there", "nested/something.md": "# Something"},
            "intermediate",
            [
                TreeNode(
                    {"bar.txt": "Something Else"}, "version it!", tag_name="1.0.0"
                ),
                TreeNode(
                    {"bar.txt": "One More"},
                    "more features",
                    [
                        TreeNode(
                            {"bar.md": "# Mark it down"}, "next rev", tag_name="1.1.0"
                        ),
                    ],
                    branch_name="my-feature",
                ),
            ],
        ),
    ],
)


@pytest.fixture(scope="session")
def repo_path():
    with tempfile.TemporaryDirectory() as workdir:
        repo_path = temp_git_repo(workdir, TEST_REPO_TREE)
        yield repo_path


## Tests #######################################################################


def test_setup_vcs_with_defaults(workdir, repo_path):
    """Make sure setup_vcs works with defaults as expected"""
    setup_vcs(repo_path)
    dest_path = os.path.join(workdir, "oper8_vcs")
    assert os.path.isdir(dest_path)
    validate_vcs_dir(repo_path, dest_path, ["1.0.0", "1.1.0"])


def test_setup_vcs_branches(workdir, repo_path):
    """Make sure that setup_vcs includes branches correctly"""
    setup_vcs(repo_path, branch_expr=[".*"], tag_expr=None)
    dest_path = os.path.join(workdir, "oper8_vcs")
    assert os.path.isdir(dest_path)
    validate_vcs_dir(repo_path, dest_path, ["my-feature", "branch/one", "main"])


def test_setup_vcs_some_of_each(workdir, repo_path):
    """Make sure that setup_vcs includes some branches and tags selectively"""
    setup_vcs(repo_path, branch_expr=["my-feature"], tag_expr=["1.1.0"])
    dest_path = os.path.join(workdir, "oper8_vcs")
    assert os.path.isdir(dest_path)
    validate_vcs_dir(repo_path, dest_path, ["my-feature", "1.1.0"])


def test_setup_vcs_force(workdir, repo_path):
    """Make sure that the dest is only overwritten if "force" is true"""
    dest_path = os.path.join(workdir, "something_else")

    # Set up something that already exists
    os.makedirs(os.path.join(dest_path))
    with open(os.path.join(dest_path, "some_file.txt"), "w") as handle:
        handle.write("Not gonna last long!")
    nested_dir = os.path.join(dest_path, "nested_dir")
    os.makedirs(nested_dir)
    with open(os.path.join(nested_dir, "some_other_file.txt"), "w") as handle:
        handle.write("bye bye")

    # Without force, it's a ValueError
    with pytest.raises(ValueError):
        setup_vcs(repo_path, destination=dest_path)

    # With force, it's fine
    setup_vcs(repo_path, destination=dest_path, force=True)
    validate_vcs_dir(repo_path, dest_path, ["1.0.0", "1.1.0"])


def test_setup_vcs_not_a_repo(workdir):
    """Make sure an error is raised if the source is not a repo"""
    with pytest.raises(ValueError):
        setup_vcs(workdir)


def test_setup_vcs_dest_is_file(workdir, repo_path):
    """Make sure that an error is raised if the destination is a file"""
    dest_path = os.path.join(workdir, "something_else")
    with open(os.path.join(dest_path), "w") as handle:
        handle.write("Not gonna last long!")
    with pytest.raises(ValueError):
        setup_vcs(repo_path, destination=dest_path, force=True)



"""
The common Connection type for a COS instance
"""

# Standard
from typing import Optional, Tuple

# First Party
import alog

# Local
from .... import Session, assert_cluster, assert_precondition
from ...utils import common
from .. import DatastoreConnectionBase

## CosConnection ###############################################################

log = alog.use_channel("COSCON")


class CosConnection(DatastoreConnectionBase):
    """
    A CosConnection holds the core connection information for a named COS
    instance, regardless of what ICosComponent implements it. The key pieces of
    information are:

    * General config
        * hostname: The hostname where the instance can be reached
        * port: The port where the instance is listening
        * bucket_name: The name of the bucket within the instance

    * Auth
        * auth_secret_name: The in-cluster name for the secret holding the
            access_key and secret_key
        * auth_secret_access_key_field: The field within the auth secret that
            holds the access_key
        * auth_secret_secret_key_field: The field within the auth secret that
            holds the secret_key

    * TLS:
        * tls_cert: The content of the TLS cert if tls is enabled
        * tls_secret_name: The in-cluster name for the secret holding the TLS
            creds if tls is enabled
        * tls_secret_cert_field: The field within the tls secret that holds the
            cert
    """

    def __init__(
        self,
        session: Session,
        hostname: str,
        port: int,
        bucket_name: str,
        auth_secret_name: str,
        auth_secret_access_key_field: str,
        auth_secret_secret_key_field: str,
        tls_secret_name: Optional[str] = None,
        tls_secret_cert_field: Optional[str] = None,
        access_key: Optional[str] = None,
        secret_key: Optional[str] = None,
        tls_cert: Optional[str] = None,
    ):
        super().__init__(session)

        # These fields must be passed in directly
        self._hostname = hostname
        self._port = port
        self._bucket_name = bucket_name
        self._auth_secret_name = auth_secret_name
        self._auth_secret_access_key_field = auth_secret_access_key_field
        self._auth_secret_secret_key_field = auth_secret_secret_key_field
        self._tls_secret_name = tls_secret_name
        self._tls_secret_cert_field = tls_secret_cert_field

        # The secret content may be populated or not, depending on whether this
        # Connection is being created by the component or a CR config. If not
        # populated now, they will be lazily populated on client request.
        self._access_key = access_key
        self._secret_key = secret_key
        self._tls_cert = tls_cert

        # Ensure that the TLS arguments are provided in a reasonable way. The
        # cert may be omitted
        tls_args = {tls_secret_name, tls_secret_cert_field}
        assert (
            tls_args == {None} or None not in tls_args
        ), "All TLS arguments must be provided if tls is enabled"
        self._tls_enabled = tls_args != {None}
        assert (
            self._tls_enabled or self._tls_cert is None
        ), "Cannot give a tls cert value when tls is disabled"

        # Schema is deduced based on the presence of the tls arguments
        self._schema = "http" if tls_secret_name is None else "https"

    ## Properties ##############################################################

    @property
    def schema(self) -> str:
        """The schema (http or https)"""
        return self._schema

    @property
    def hostname(self) -> str:
        """The hostname (without schema)"""
        return self._hostname

    @property
    def port(self) -> int:
        """The numeric port"""
        return self._port

    @property
    def endpoint(self) -> int:
        """The fully constructed endpoint for the COS instance"""
        return f"{self._schema}://{self._hostname}:{self._port}"

    @property
    def bucket_name(self) -> int:
        """The numeric bucket_name"""
        return self._bucket_name

    @property
    def auth_secret_name(self) -> str:
        """Secret name containing the access_key and secret_key"""
        return self._auth_secret_name

    @property
    def auth_secret_access_key_field(self) -> str:
        """Field in the auth secret containing the access_key"""
        return self._auth_secret_access_key_field

    @property
    def auth_secret_secret_key_field(self) -> str:
        """Field in the auth secret containing the secret_key"""
        return self._auth_secret_secret_key_field

    @property
    def tls_secret_name(self) -> str:
        """The name of the secret holding the tls certificate (for mounting)"""
        return self._tls_secret_name

    @property
    def tls_secret_cert_field(self) -> str:
        """The field within the tls secret that holds the CA cert"""
        return self._tls_secret_cert_field

    @property
    def tls_enabled(self) -> bool:
        return self._tls_enabled

    ## Interface ###############################################################

    _DICT_FIELDS = [
        "hostname",
        "port",
        "bucket_name",
        "auth_secret_name",
        "auth_secret_access_key_field",
        "auth_secret_secret_key_field",
        "tls_secret_name",
        "tls_secret_cert_field",
    ]

    def to_dict(self) -> dict:
        """Return the dict representation of the object for the CR"""
        return {field: getattr(self, f"_{field}") for field in self._DICT_FIELDS}

    @classmethod
    def from_dict(cls, session: Session, config_dict: dict) -> "CosConnection":
        kwargs = {"session": session}
        config_dict = common.camelcase_to_snake_case(config_dict)

        uri_secret = config_dict.get("uri_secret")
        uri_hostname_field = config_dict.get("uri_secret_hostname_field")
        uri_port_field = config_dict.get("uri_secret_port_field")
        uri_bucketname_field = config_dict.get("uri_secret_bucketname_field")

        # First pull provided hostname/port secret if available and fill in
        # hostname/port fields into config_dict
        if (
            uri_secret
            and uri_hostname_field
            and uri_port_field
            and uri_bucketname_field
        ):
            # If we have provided host/port credentials, we need to extract them
            # and place these values in our config dict
            success, secret_content = session.get_object_current_state(
                "Secret", uri_secret
            )
            assert_cluster(success, f"Fetching connection secret [{uri_secret}] failed")
            assert "data" in secret_content, "Got a secret without 'data'?"
            secret_content = secret_content.get("data", {})
            assert_precondition(
                secret_content,
                f"Missing expected Secret/{uri_secret} holding [hostname] and [port]",
            )
            hostname = common.b64_secret_decode(secret_content.get(uri_hostname_field))
            port = common.b64_secret_decode(secret_content.get(uri_port_field))
            bucketname = common.b64_secret_decode(
                secret_content.get(uri_bucketname_field)
            )
            if None in [hostname, port, bucketname]:
                log.debug2(
                    "Failed to find hostname/port/bucketname in uri secret [%s]",
                    uri_secret,
                )

            config_dict["hostname"], config_dict["port"], config_dict["bucket_name"] = (
                hostname,
                int(port),
                bucketname,
            )

        for field in cls._DICT_FIELDS:
            if field not in config_dict:
                raise ValueError(f"Missing required connection element [{field}]")

            # Set the kwargs (using None in place of empty strings)
            kwargs[field] = config_dict[field] or None
        return cls(**kwargs)

    ## Client Utilities ########################################################

    def get_auth_keys(self) -> Tuple[Optional[str], Optional[str]]:
        """Get the current access_key/secret_key pair from the auth secret if
        available

        Returns:
            access_key:  str or None
                The plain-text access_key (not encoded) if available
            secret_key:  str or None
                The plain-text secret_key (not encoded) if available
        """
        if None in [self._access_key, self._secret_key]:
            secret_content = self._fetch_secret_data(self._auth_secret_name) or {}
            log.debug4("Auth secret content: %s", secret_content)
            log.debug3(
                "Looking for [%s/%s]",
                self._auth_secret_access_key_field,
                self._auth_secret_secret_key_field,
            )
            access_key = secret_content.get(self._auth_secret_access_key_field)
            secret_key = secret_content.get(self._auth_secret_secret_key_field)
            if None in [access_key, secret_key]:
                log.debug2(
                    "Failed to find access_key/secret_key in auth secret [%s]",
                    self._auth_secret_name,
                )
                return None, None
            self._access_key = access_key
            self._secret_key = secret_key
        return self._access_key, self._secret_key

    def get_tls_cert(self) -> Optional[str]:
        """Get the current TLS certificate for a client connection if TLS is
        enabled

        If TLS is enabled, but the cert is not found, this function will raise
        an AssertionError

        Returns:
            tls_cert: str or None
                PEM encoded cert string (not base64-encoded) if found, otherwise
                None
        """
        if self._tls_enabled:
            if self._tls_cert is None:
                secret_data = self._fetch_secret_data(self._tls_secret_name)
                if secret_data is not None:
                    self._tls_cert = secret_data.get(self._tls_secret_cert_field)
            return self._tls_cert

        return None

    def get_connection_string(self) -> str:
        """Get the formatted s3 connection string to connect to the given bucket
        in the instance

        Returns:
            connection_string:  str
                The formatted connection string
        """
        access_key, secret_key = self.get_auth_keys()
        assert_precondition(
            None not in [access_key, secret_key],
            "No auth keys available for COS connection string",
        )
        return (
            "s3,endpoint={}://{}:{},accesskey={},secretkey={},bucketsuffix={}".format(
                self._schema,
                self._hostname,
                self._port,
                access_key,
                secret_key,
                self._bucket_name,
            )
        )



"""
This module uses VCS to create a trimmed down repo with a selection of local
branches and tags and a fully flattened history.
"""

# Standard
from typing import List, Optional
import os
import re
import shutil

# First Party
import alog

# Local
from .vcs import VCS, VCSConfigError

## Globals #####################################################################

log = alog.use_channel("VCSSU")

# Sentinel so that None can be passed to tag_expr
DEFAULT_TAG_EXPR = r"[0-9]+\.[0-9]+\.[0-9]+"
__UNSET__ = "__UNSET__"

# Default value for the destination
DEFAULT_DEST = "oper8_vcs"

## Public ######################################################################


def setup_vcs(
    source: str,
    destination: Optional[str] = None,
    branch_expr: Optional[List[str]] = None,
    tag_expr: Optional[List[str]] = __UNSET__,
    force: bool = False,
):
    """This utility will initialize an operator's VCS directory for use with
    oper8's VCS versioning.

    Args:
        source (str): The path to the source repository on disk
        destination (Optional[str]): The path where the VCS repo should be
            created
        branch_expr (Optional[List[str]]): Regular expression(s) to use to
            identify branches to retain in the VCS repo
        tag_expr (Optional[List[str]]): Regular expression(s) to use to
            identify tags to retain in the VCS repo
        force (bool): Force overwrite existing destination
    """
    initializer = VCSRepoInitializer(
        source=source, destination=destination or DEFAULT_DEST, force=force
    )
    initializer.initialize_branches(
        branch_expr=branch_expr,
        tag_expr=tag_expr if tag_expr is not __UNSET__ else [DEFAULT_TAG_EXPR],
    )
    initializer.clean_up()


## Impl ########################################################################


class VCSRepoInitializer:
    """This class encapsulates the key attributes of the VCS repo initialization
    process
    """

    # The git repo that is being compressed for VCS versioning
    source_repo: VCS
    # The git repo where the VCS versioning repo is going to be created
    dest_repo: VCS
    # The remote within the destination repo that refers to the source repo
    source_remote: str
    # The reference to the root empty commit in the destination repo
    root_ref: str

    # Default branch name used when creating the repo
    DEFAULT_BRANCH_NAME = "__root__"

    # The name of the source remote
    SOURCE_REMOTE = "__source__"

    def __init__(self, source: str, destination: str, force: bool):
        """Initialize and set up the repos and common attributes"""

        # Make sure the source is a git repo
        try:
            self.source_repo = VCS(source)
        except VCSConfigError as err:
            msg = f"Invalid source git repo: {source}"
            log.error(msg)
            raise ValueError(msg) from err
        log.debug("Source Repo: %s", source)

        # Set up the dest and make sure it's empty
        if os.path.isfile(destination):
            msg = f"Invalid destination: {destination} is a file"
            log.error(msg)
            raise ValueError(msg)
        os.makedirs(destination, exist_ok=True)
        contents = os.listdir(destination)
        if contents:
            if not force:
                msg = f"Invalid destination: {destination} is not empty"
                log.error(msg)
                raise ValueError(msg)
            log.debug("Force cleaning dest %s", destination)
            for entry in contents:
                full_path = os.path.join(destination, entry)
                if os.path.isdir(full_path):
                    log.debug3("Removing dir: %s", full_path)
                    shutil.rmtree(full_path)
                else:
                    log.debug3("Removing file: %s", full_path)
                    os.remove(full_path)

        # Initialize the dest as an empty repo
        log.info("Initializing dest repo: %s", destination)
        self.dest_repo = VCS(
            destination, create_if_needed=True, initial_head=self.DEFAULT_BRANCH_NAME
        )
        self.dest_repo.create_commit("root")
        self.dest_repo.add_remote(self.SOURCE_REMOTE, source)
        self.root_ref = self.dest_repo.head

    def initialize_branches(
        self,
        branch_expr: Optional[List[str]],
        tag_expr: Optional[List[str]],
    ):
        """Perform the initialize of all branches in the destination repo from
        the branches and tags that match the given expressions.
        """
        # Get all tags and branches
        tags = self._list_tags(self.source_repo)
        branches = self._list_branches(self.source_repo)
        log.debug2("All Tags: %s", tags)
        log.debug2("All Branches: %s", branches)

        # Filter the tags and branches by the filter arguments
        keep_tags = self._filter_refs(tags, tag_expr)
        keep_branches = self._filter_refs(branches, branch_expr)
        log.debug2("Keep Tags: %s", keep_tags)
        log.debug2("Keep Branches: %s", keep_branches)

        # For each retained reference, fetch the ref from the source, check out the
        # files to the dest, and make a fresh commit
        for keep_ref in keep_tags:
            log.debug("Making destination branch [%s] from tag", keep_ref)
            self._make_dest_branch(keep_ref, False)
        for keep_ref in keep_branches:
            log.debug("Making destination branch [%s] from branch", keep_ref)
            self._make_dest_branch(keep_ref, True)

    def clean_up(self):
        """Clean out all unnecessary content from the destination repo"""

        # Check the root back out
        log.debug3("Checking out root")
        self.dest_repo.checkout_ref(self.root_ref)

        # Delete the source remote
        self.dest_repo.delete_remote(self.SOURCE_REMOTE)

        # Remove all tags
        for tag_name in self._list_tags(self.dest_repo):
            self.dest_repo.delete_tag(tag_name)

        # Remove the root branch and leave HEAD detached
        self.dest_repo.checkout_detached_head()
        self.dest_repo.delete_branch(self.DEFAULT_BRANCH_NAME)

        # Compress the references to remove orphaned refs and objects
        self.dest_repo.compress_references()

    ## Impl ##

    def _get_all_checkout_files(self, keep_ref: str) -> List[str]:
        """Get all of the file paths in the given ref relative to the dest repo

        # NOTE: This relies on pygit2 syntax!
        """
        commit, _ = self.dest_repo.get_ref(keep_ref)
        diff = commit.tree.diff_to_workdir()
        return [delta.new_file.path for delta in diff.deltas]

    def _make_dest_branch(self, keep_ref: str, is_branch: bool):
        """This is the function that does the main work of copying code from the
        source to the destination and creating a clean commit.
        """
        # Make sure the root is checked out in the destination repo
        log.debug3("Checking out root")
        self.dest_repo.checkout_ref(self.root_ref)

        # Fetch the ref to keep from the source in the dest
        log.debug3("Fetching %s", keep_ref)
        self.dest_repo.fetch_remote(self.SOURCE_REMOTE, {keep_ref})

        # Check out the files
        remote_ref_name = keep_ref
        if is_branch:
            remote_ref_name = f"refs/remotes/{self.SOURCE_REMOTE}/{keep_ref}"
        log.debug3("Checking out files for %s", remote_ref_name)
        self.dest_repo.checkout_ref(
            remote_ref_name, paths=self._get_all_checkout_files(remote_ref_name)
        )

        # Make a new branch named with this ref's shorthand name with any remote
        # information removed
        branch_name = keep_ref
        log.debug2("Dest branch name: %s", branch_name)
        root_commit, _ = self.dest_repo.get_ref(self.root_ref)
        branch = self.dest_repo.create_branch(branch_name, root_commit)
        self.dest_repo.checkout_ref(branch.name)

        # Make a commit with these files
        self.dest_repo.create_commit(keep_ref, parents=[root_commit.oid])

        # Check the root branch back out
        self.dest_repo.checkout_ref(self.DEFAULT_BRANCH_NAME)

    ## Static Helpers ##

    @staticmethod
    def _list_branches(repo: VCS) -> List[str]:
        """List all of the local branches

        Args:
            repo (VCS): The repo to list

        Returns
            refs (List[str]): A set of all branch references
        """
        refs = set()
        for ref in repo.list_refs():
            _, repo_ref = repo.get_ref(ref)
            name_parts = repo_ref.name.split("/")
            if (
                "tags" not in name_parts
                and "HEAD" not in name_parts
                and "remotes" not in name_parts
                and repo_ref.name != "refs/stash"
            ):
                refs.add(ref)
        return sorted(sorted(refs))

    @staticmethod
    def _list_tags(repo: VCS) -> List[str]:
        """List all of the tags and references in the repo

        Args:
            repo (VCS): The repo to list

        Returns
            refs (List[str]): A set of all tag references
        """
        return list(
            sorted(
                {
                    ref
                    for ref in repo.list_refs()
                    if "refs/tags" in repo.get_ref(ref)[1].name
                }
            )
        )

    @staticmethod
    def _filter_refs(refs: List[str], exprs: Optional[List[str]]) -> List[str]:
        """Keep all refs that match at least one of the expressions"""
        return [ref for ref in refs if any(re.match(expr, ref) for expr in exprs or [])]



"""Standard data types used through PWM"""

# Standard
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from functools import cached_property
from multiprocessing.connection import Connection
from typing import Dict, List, NamedTuple, Type, Union
import abc
import importlib
import multiprocessing

# Local
from ....managed_object import ManagedObject

### General Use Data Classes

# Forward Declarations
CONTROLLER_TYPE = "Controller"
FILTER_MANAGER_TYPE = "FilterManager"
FILTER_TYPE = "Filter"
KUBE_EVENT_TYPE_TYPE = "KubeEventType"


@dataclass(eq=True, frozen=True)
class ResourceId:
    """Class containing the information needed to identify a resource"""

    api_version: str
    kind: str
    name: str = None
    namespace: str = None

    # Id properties

    @cached_property
    def global_id(self) -> str:
        """Get the global_id for a resource in the form kind.version.group"""
        group_version = self.api_version.split("/")
        return ".".join([self.kind, *reversed(group_version)])

    @cached_property
    def namespaced_id(self) -> str:
        """Get the namespace specific id for a resource"""
        return f"{self.namespace}.{self.global_id}"

    # Helper Accessor functions
    def get_id(self) -> str:
        """Get the requisite id for a resource"""
        return self.namespaced_id if self.namespace else self.global_id

    def get_named_id(self) -> str:
        """Get a named id for a resouce"""
        return f"{self.name}.{self.get_id()}"

    def get_resource(self) -> dict:
        """Get a resource template from this id"""
        return {
            "kind": self.kind,
            "apiVersion": self.api_version,
            "metadata": {"name": self.name, "namespace": self.namespace},
        }

    # Helper Creation Functions
    @classmethod
    def from_resource(cls, resource: Union[ManagedObject, dict]) -> "ResourceId":
        """Create a resource id from an existing resource"""
        metadata = resource.get("metadata", {})
        return cls(
            api_version=resource.get("apiVersion"),
            kind=resource.get("kind"),
            namespace=metadata.get("namespace"),
            name=metadata.get("name"),
        )

    @classmethod
    def from_owner_ref(cls, owner_ref: dict, namespace: str = None) -> "ResourceId":
        """Create a resource id from an ownerRef"""
        return cls(
            api_version=owner_ref.get("apiVersion"),
            kind=owner_ref.get("kind"),
            namespace=namespace,
            name=owner_ref.get("name"),
        )

    @classmethod
    def from_controller(
        cls, controller: Type[CONTROLLER_TYPE], namespace: str = None
    ) -> "ResourceId":
        """Get a Controller's target as a resource id"""
        return cls(
            api_version=f"{controller.group}/{controller.version}",
            kind=controller.kind,
            namespace=namespace,
        )


### Watch Data Classes


class ClassInfo(NamedTuple):
    """Class containing information describing a class. This is required when passing class
    references between processes which might have different sys paths like when using VCS"""

    moduleName: str
    className: str

    # Generation Utilities
    @classmethod
    def from_type(cls, class_obj: type) -> "ClassInfo":
        """Create a ClassInfo from a class object"""
        return cls(moduleName=class_obj.__module__, className=class_obj.__name__)

    @classmethod
    def from_obj(cls, obj) -> "ClassInfo":
        """Create a ClassInfo from an existing object"""
        return cls.from_type(type(obj))

    # Get the class referenced described by the info
    def to_class(self) -> type:
        """Import and return a ClassInfo's type"""
        module = importlib.import_module(self.moduleName)
        if not module:
            raise ValueError(f"Invalid ControllerInfo Module: {self.moduleName}")

        if not hasattr(module, self.className):
            raise ValueError(
                f"Invalid ControllerInfo: {self.className} not a member of {self.moduleName}"
            )

        return getattr(module, self.className)


@dataclass
class WatchedResource:
    """A class for tracking a resource in the cluster. Every resource that has a
    requested watch will have a corresponding WatchedResource"""

    gvk: ResourceId
    # Each watched resource contains a dict of filters for each
    # corresponding watch request. The key is the named_id of
    # the requester or None for default filters. This aligns with
    # the Controllers pwm_filters attribute
    filters: Dict[str, FILTER_MANAGER_TYPE] = field(default_factory=dict)


@dataclass()
class WatchRequest:
    """A class for requesting a watch of a particular object. It contains information around the
    watched object, who requested the watch, the controller type to be reconciled, and any filters
    to be applied to just this request"""

    watched: ResourceId
    requester: ResourceId

    # Watch request must have either type or info
    controller_type: Type[CONTROLLER_TYPE] = None
    controller_info: ClassInfo = None

    # Don't compare filters when checking equality as we
    # assume they're the same if they have the same controller
    filters: List[Type[FILTER_TYPE]] = field(default_factory=list, compare=False)
    filters_info: List[Type[ClassInfo]] = field(default_factory=list, compare=False)

    def __hash__(self) -> int:
        return hash(
            (
                self.watched,
                self.requester,
                self.controller_type if self.controller_type else self.controller_info,
            )
        )


##  Reconcile Enums


class ReconcileRequestType(Enum):
    """Enum to expand the possible KubeEventTypes to include PythonWatchManager
    specific events"""

    # Used for events that are a requeue of an object
    REQUEUED = "REQUEUED"

    # Used for periodic reconcile events
    PERIODIC = "PERIODIC"

    # Used for when an event is a dependent resource of a controller
    DEPENDENT = "DEPENDENT"

    # Used as a sentinel to alert threads to stop
    STOPPED = "STOPPED"


### Reconcile Classes


@dataclass
class ReconcileRequest:
    """Class to represent one request to the ReconcileThread. This includes
    important information including the current resource and Controller being
    reconciled.
    """

    controller_type: Type[CONTROLLER_TYPE]
    type: Union[ReconcileRequestType, KUBE_EVENT_TYPE_TYPE]
    resource: ManagedObject
    timestamp: datetime = datetime.now()

    def uid(self):
        """Get the uid of the resource being reconciled"""
        return self.resource.uid


@dataclass
class ReconcileProcess:
    """Dataclass to track a running reconcile. This includes the raw process
    object, the result pipe, and the request being reconciled"""

    process: multiprocessing.Process
    request: ReconcileRequest
    pipe: Connection

    def fileno(self):
        """Pass through fileno to process. Sentinel so this object can be
        directly used by multiprocessing.connection.wait"""
        return self.process.sentinel

    def uid(self):
        """Get the uid for the resource being reconciled"""
        return self.request.uid()


### Timer Data Classes
@dataclass(order=True)
class TimerEvent:
    """Class for keeping track of an item in the timer queue. Time is the
    only comparable field to support the TimerThreads priority queue"""

    time: datetime
    action: callable = field(compare=False)
    args: list = field(default_factory=list, compare=False)
    kwargs: dict = field(default_factory=dict, compare=False)
    stale: bool = field(default=False, compare=False)

    def cancel(self):
        """Cancel this event. It will not be executed when read from the
        queue"""
        self.stale = True


## Meta Classes


class Singleton(type):
    """MetaClass to limit a class to only one global instance. When the
    first instance is created it's attached to the Class and the next
    time someone initializes the class the original instance is returned
    """

    def __call__(cls, *args, **kwargs):
        if getattr(cls, "_disable_singleton", False):
            return type.__call__(cls, *args, **kwargs)

        # The _instance is attached to the class itself without looking upwards
        # into any parent classes
        if "_instance" not in cls.__dict__:
            cls._instance = type.__call__(cls, *args, **kwargs)
        return cls._instance


class ABCSingletonMeta(Singleton, abc.ABCMeta):
    """Shared metaclass for ABCMeta and Singleton"""



"""
This module holds the common functionality used to represent the status of
resources managed by oper8

Oper8 supports the following orthogonal status conditions:

* Ready: True if the service is able to take traffic
* Updating: True if a modification is being actively applied to the application

Additionally, oper8 supports a top-level status element to report the detailed
status of the managed components. The schema is:
{
    "componentStatus": {
        "allComponents": [list of all node names],
        "deployedComponents": [list of nodes that have successfully deployed],
        "verifiedComponents": [list of nodes that have successfully verified],
        "failedComponents": [list of nodes that have successfully verified],
        "deployed": "N/M",
        "verified": "N/M",
    }
}
"""

# Standard
from datetime import datetime
from enum import Enum
from typing import List, Optional, Union
import copy

# Third Party
from deepdiff import DeepDiff

# First Party
import alog

# Local
from . import config
from .dag import CompletionState
from .utils import nested_get, nested_set

log = alog.use_channel("STTUS")

## Public ######################################################################

# The "type" values in the condition
READY_CONDITION = "Ready"
UPDATING_CONDITION = "Updating"

# The key in the condition used for the timestamp
TIMESTAMP_KEY = "lastTransactionTime"

# The keys for component status information
COMPONENT_STATUS = "componentStatus"
COMPONENT_STATUS_ALL_NODES = "allComponents"
COMPONENT_STATUS_DEPLOYED_NODES = "deployedComponents"
COMPONENT_STATUS_UNVERIFIED_NODES = "unverifiedComponents"
COMPONENT_STATUS_FAILED_NODES = "failedComponents"
COMPONENT_STATUS_DEPLOYED_COUNT = "deployed"
COMPONENT_STATUS_VERIFIED_COUNT = "verified"

# The fields in status that hold the version information
# NOTE: These intentionally match IBM CloudPak naming conventions
VERSIONS_FIELD_CURRENT_VERSION = "versions.reconciled"
VERSIONS_FIELD_AVAILABLE_VERSIONS = "versions.available.versions"
# The field in status that holds the operator version information
OPERATOR_VERSION = "operatorVersion"


class ReadyReason(Enum):
    """Nested class to hold reason constants for the Ready condition"""

    # The application is stable and ready for traffic
    STABLE = "Stable"

    # The application is rolling out for the first time
    INITIALIZING = "Initializing"

    # The application rollout is in progress and will continue
    # the next reconcile
    IN_PROGRESS = "InProgress"

    # The application has hit an unrecoverable config error during rollout
    CONFIG_ERROR = "ConfigError"

    # The application has hit an unrecoverable error during rollout
    ERRORED = "Errored"


class UpdatingReason(Enum):
    """Nested class to hold reason constants for the Updating condition"""

    # There are no updates to apply to the application
    STABLE = "Stable"

    # A required precondition was not met
    PRECONDITION_WAIT = "PreconditionWait"

    # A required deployment verification condition was not met
    VERIFY_WAIT = "VerifyWait"

    # The application attempted to perform an operation against the cluster that
    # failed unexpectedly
    CLUSTER_ERROR = "ClusterError"

    # An error occurred, so the application is not attempting to update
    ERRORED = "Errored"

    # Version upgrade is initiated
    VERSION_CHANGE = "VersionChange"


class ServiceStatus(Enum):
    """Nested class to hold status constants for the service status"""

    # Installation or Update reconciliation is in-progress
    IN_PROGRESS = "InProgress"

    # Installation or Update failed with error
    FAILED = "Failed"

    # Service is in stable state
    COMPLETED = "Completed"


def make_application_status(  # pylint: disable=too-many-arguments,too-many-locals
    ready_reason: Optional[Union[ReadyReason, str]] = None,
    ready_message: str = "",
    updating_reason: Optional[Union[UpdatingReason, str]] = None,
    updating_message: str = "",
    component_state: Optional[CompletionState] = None,
    external_conditions: Optional[List[dict]] = None,
    external_status: Optional[dict] = None,
    version: Optional[str] = None,
    supported_versions: Optional[List[str]] = None,
    operator_version: Optional[str] = None,
    kind: Optional[str] = None,
) -> dict:
    """Create a full status object for an application

    Args:
        ready_reason:  Optional[ReadyReason or str]
            The reason enum for the Ready condition
        ready_message:  str
            Plain-text message explaining the Ready condition value
        updating_reason:  Optional[UpdatingReason or str]
            The reason enum for the Updating condition
        updating_message:  str
            Plain-text message explaining the Updating condition value
        component_state:  Optional[CompletionState]
            The terminal state of components in the latest rollout
        external_conditions:  Optional[List[dict]]
            Additional conditions to include in the update
        external_status:  Optional[dict]
            Additional key/value status elements besides "conditions" that
            should be preserved through the update
        version:  Optional[str]
            The verified version of the application
        supported_versions:  Optional[List[str]]
            The list of supported versions for this application
        operator_version:  Optional[str]
            The operator version for this application
        kind: Optional[str]
            The kind of reconciliing CR. If specified, this function adds
            service status field which is compliant with IBM Cloud Pak
            requirements.

    Returns:
        current_status:  dict
            Dict representation of the status for the application
    """
    now = datetime.now()
    conditions = []
    if ready_reason is not None:
        conditions.append(_make_ready_condition(ready_reason, ready_message, now))
    if updating_reason is not None:
        conditions.append(
            _make_updating_condition(updating_reason, updating_message, now)
        )
    conditions.extend(external_conditions or [])
    status = external_status or {}
    status["conditions"] = conditions

    # If a component_state is given, create the top-level status elements to
    # track which components have deployed and verified
    if component_state is not None:
        log.debug2("Adding component state to status")
        status[COMPONENT_STATUS] = _make_component_state(component_state)
        log.debug3(status[COMPONENT_STATUS])

    # Create the versions section
    if version is not None:
        nested_set(status, VERSIONS_FIELD_CURRENT_VERSION, version)
    if supported_versions is not None:
        nested_set(
            status,
            VERSIONS_FIELD_AVAILABLE_VERSIONS,
            [_make_available_version(version) for version in supported_versions],
        )
    if operator_version is not None:
        nested_set(status, OPERATOR_VERSION, operator_version)

    # Create service status section
    if kind:
        # make field name follow k8s naming convention
        service_status_field = kind[0].lower()
        if len(kind) > 1:
            service_status_field += kind[1:]
        service_status_field += "Status"

        # Only update service status if the current value is set by oper8. This
        # allows services to override the service status section
        current_service_status = status.get(service_status_field)
        managed_service_values = [status.value for status in ServiceStatus]
        if (
            not current_service_status
            or current_service_status in managed_service_values
        ):
            current_service_status = _make_service_status(
                ready_reason, updating_reason
            ).value

        status[service_status_field] = current_service_status

    return status


def update_application_status(current_status: dict, **kwargs) -> dict:
    """Create an updated status based on the values in the current status

    Args:
        current_status:  dict
            The dict representation of the status for a given application
        **kwargs:
            Additional keyword args to pass to make_application_status

    Returns:
        updated_status:  dict
            Updated dict representation of the status for the application
    """
    # Make a deep copy of current_status so that we aren't accidentally
    # modifying the current status object. This prevents a bug where status
    # changes are not detected
    current_status = copy.deepcopy(current_status)

    # Make a dict of type -> condition. This is necessary because other
    # conditions may be applied by ansible
    current_conditions = current_status.get("conditions", [])
    current_condition_map = {cond["type"]: cond for cond in current_conditions}
    ready_cond = current_condition_map.get(READY_CONDITION, {})
    updating_cond = current_condition_map.get(UPDATING_CONDITION, {})

    # Setup the kwargs for the status call
    ready_reason = ready_cond.get("reason")
    updating_reason = updating_cond.get("reason")
    if ready_reason:
        kwargs.setdefault("ready_reason", ReadyReason(ready_reason))
    if updating_reason:
        kwargs.setdefault("updating_reason", UpdatingReason(updating_reason))
    kwargs.setdefault("ready_message", ready_cond.get("message", ""))
    kwargs.setdefault("updating_message", updating_cond.get("message", ""))

    # Extract external conditions managed by other portions of the operator
    external_conditions = [
        cond
        for cond in current_conditions
        if cond.get("type") not in [READY_CONDITION, UPDATING_CONDITION]
    ]
    log.debug3("External conditions: %s", external_conditions)
    kwargs["external_conditions"] = external_conditions
    log.debug3("Merged status kwargs: %s", kwargs)

    # Extract external status elements (besides conditions) managed by other
    # portions of the operator
    external_status = {
        key: val for key, val in current_status.items() if key != "conditions"
    }
    kwargs["external_status"] = external_status
    kwargs["operator_version"] = config.operator_version

    return make_application_status(**kwargs)


def update_resource_status(
    deploy_manager: "DeployManagerBase",  # noqa: F821
    kind: str,
    api_version: str,
    name: str,
    namespace: str,
    **kwargs: dict,
) -> dict:
    """Create an updated status based on the values in the current status

    Args:
        deploy_manager: DeployManagerBase
            The deploymanager used to get and set status
        kind: str
            The kind of the resource
        api_version: str
            The api_version of the resource
        name: str
            The name of the resource
        namespace: str
            The namespace the resource is located in
        **kwargs: Dict
            Any additional keyword arguments to be passed to update_application_status

    Returns:
        status_object: Dict
            The applied status if successful

    """
    log.debug3(
        "Updating status for %s/%s.%s/%s",
        namespace,
        api_version,
        kind,
        name,
    )

    # Fetch the current status from the cluster
    success, current_state = deploy_manager.get_object_current_state(
        api_version=api_version,
        kind=kind,
        name=name,
        namespace=namespace,
    )
    if not success:
        log.warning("Failed to fetch current state for %s/%s/%s", namespace, kind, name)
        return {}
    current_status = (current_state or {}).get("status", {})
    log.debug3("Pre-update status: %s", current_status)

    # Merge in the given status
    status_object = update_application_status(current_status, kind=kind, **kwargs)
    log.debug3("Updated status: %s", status_object)

    # Check to see if the status values of any conditions have changed and
    # only update the status if it has changed
    if status_changed(current_status, status_object):
        log.debug("Found meaningful change. Updating status")
        log.debug2("(current) %s != (updated) %s", current_status, status_object)

        # Do the update
        success, _ = deploy_manager.set_status(
            kind=kind,
            name=name,
            namespace=namespace,
            api_version=api_version,
            status=status_object,
        )

        # Since this is just a status update, we don't fail if the update fails,
        # but we do throw a warning
        if not success:
            log.warning("Failed to update status for [%s/%s/%s]", namespace, kind, name)
            return {}

    return status_object


def status_changed(current_status: dict, new_status: dict) -> bool:
    """Compare two status objects to determine if there is a meaningful change
    between the current status and the proposed new status. A meaningful change
    is defined as any change besides a timestamp.

    Args:
        current_status:  dict
            The raw status dict from the current CR
        new_status:  dict
            The proposed new status

    Returns:
        status_changed:  bool
            True if there is a meaningful change between the current status and
            the new status
    """
    # Status objects must be dicts
    if not isinstance(current_status, dict) or not isinstance(new_status, dict):
        return True

    # Perform a deep diff, excluding timestamps
    return bool(
        DeepDiff(
            current_status,
            new_status,
            exclude_obj_callback=lambda _, path: path.endswith(f"{TIMESTAMP_KEY}']"),
        )
    )


def get_condition(type_name: str, current_status: dict) -> dict:
    """Extract the given condition type from a status object

    Args:
        type:  str
            The condition type to fetch
        current_status:  dict
            The dict representation of the status for a given application

    Returns:
        condition:  dict
            The condition object if found, empty dict otherwise
    """
    cond = [
        cond
        for cond in current_status.get("conditions", [])
        if cond.get("type") == type_name
    ]
    if cond:
        assert len(cond) == 1, f"Found multiple condition entries for {type_name}"
        return cond[0]
    return {}


def get_version(current_status: dict) -> Optional[str]:
    """Extract the current version (not desired version) from a status object

    Args:
        current_status: dict
            The dict representation of the status for a given application

    Returns:
        version: Optional[dict]
            The current version if found in a status object, None otherwise.

    """
    return nested_get(current_status, VERSIONS_FIELD_CURRENT_VERSION)


## Implementation Details ######################################################


def _make_status_condition(
    type_name: str,
    status: bool,
    reason: str,
    message: str,
    last_transaction_time: datetime,
):
    """Convert the condition to the dict representation to be added to the
    kubernetes object
    """
    return {
        "type": type_name,
        "status": str(status),
        "reason": reason.value,
        "message": message,
        TIMESTAMP_KEY: last_transaction_time.isoformat(),
    }


def _make_ready_condition(
    reason: Union[ReadyReason, str],
    message: str,
    last_transaction_time: datetime,
):
    """Construct a ready condition with a reason and determine the status based
    on the reason
    """
    if isinstance(reason, str):
        reason = ReadyReason(reason)
    ready_status = reason == ReadyReason.STABLE
    log.debug2("%s status %s: %s", READY_CONDITION, ready_status, reason)
    return _make_status_condition(
        READY_CONDITION, ready_status, reason, message, last_transaction_time
    )


def _make_updating_condition(
    reason: Union[UpdatingReason, str],
    message: str,
    last_transaction_time: datetime,
):
    """Construct an updating condition with a reason and determine the status
    based on the reason
    """
    if isinstance(reason, str):
        reason = UpdatingReason(reason)
    if reason in [
        UpdatingReason.STABLE,
        UpdatingReason.CLUSTER_ERROR,
        UpdatingReason.ERRORED,
    ]:
        updating_status = False
    else:
        updating_status = True

    log.debug2("%s status %s: %s", UPDATING_CONDITION, updating_status, reason)
    return _make_status_condition(
        UPDATING_CONDITION, updating_status, reason, message, last_transaction_time
    )


def _make_component_state(component_state: CompletionState) -> dict:
    """Make the component state object"""
    all_nodes = sorted([comp.get_name() for comp in component_state.all_nodes])
    deployed_nodes = sorted(
        [
            comp.get_name()
            for comp in component_state.verified_nodes.union(
                component_state.unverified_nodes
            )
        ]
    )
    verified_nodes = sorted(
        [comp.get_name() for comp in component_state.verified_nodes]
    )
    unverified_nodes = sorted(
        [comp.get_name() for comp in component_state.unverified_nodes]
    )
    failed_nodes = sorted([comp.get_name() for comp in component_state.failed_nodes])
    return {
        COMPONENT_STATUS_ALL_NODES: all_nodes,
        COMPONENT_STATUS_DEPLOYED_NODES: deployed_nodes,
        COMPONENT_STATUS_UNVERIFIED_NODES: unverified_nodes,
        COMPONENT_STATUS_FAILED_NODES: failed_nodes,
        COMPONENT_STATUS_DEPLOYED_COUNT: f"{len(deployed_nodes)}/{len(all_nodes)}",
        COMPONENT_STATUS_VERIFIED_COUNT: f"{len(verified_nodes)}/{len(all_nodes)}",
    }


def _make_available_version(version: str) -> dict:
    """Make an object for the available version list following the IBM CloudPak
    spec
    """
    return {"name": version}


def _make_service_status(
    ready_reason: Optional[Union[ReadyReason, str]] = None,
    updating_reason: Optional[Union[UpdatingReason, str]] = None,
) -> ServiceStatus:
    """Make service status based on current ready and updating reason"""
    # consider empty ready/updating status as reconciliation is starting
    if ready_reason is None or updating_reason is None:
        return ServiceStatus.IN_PROGRESS

    ready_reason = (
        ReadyReason(ready_reason) if isinstance(ready_reason, str) else ready_reason
    )
    updating_reason = (
        UpdatingReason(updating_reason)
        if isinstance(updating_reason, str)
        else updating_reason
    )

    # Completed case
    if ready_reason == ReadyReason.STABLE and updating_reason == UpdatingReason.STABLE:
        return ServiceStatus.COMPLETED

    # Failed case
    if ready_reason in [ReadyReason.ERRORED, ReadyReason.CONFIG_ERROR]:
        return ServiceStatus.FAILED
    if updating_reason in [UpdatingReason.ERRORED, UpdatingReason.CLUSTER_ERROR]:
        return ServiceStatus.FAILED

    # The other cases are considered In-progress
    return ServiceStatus.IN_PROGRESS



"""
ReconcileProcessEntrypoint for all PWM reconciles
"""

# Standard
from multiprocessing.connection import Connection
from typing import Type
import copy
import dataclasses
import logging
import multiprocessing
import os
import signal
import sys
import threading

# First Party
import aconfig
import alog

# Local
from ... import config
from ...controller import Controller
from ...deploy_manager import DeployManagerBase, KubeEventType, OpenshiftDeployManager
from ...reconcile import ReconcileManager
from .filters import (
    AndFilter,
    DependentWatchFilter,
    FilterManager,
    get_filters_for_resource_id,
)
from .utils import (
    ClassInfo,
    LogQueueHandler,
    ReconcileRequest,
    ResourceId,
    WatchRequest,
)

log = alog.use_channel("ENTRY")


## ReconcileProcessEntrypoint Deploy Manager


class ReconcileProcessDeployManager(OpenshiftDeployManager):
    """ReconcileProcessEntrypoint deploy manager is a helper deploy manager
    that allows the PWM to insert functionality during a reconcile. This
    is used for things like watching dependent resources and subsystem rollout"""

    def __init__(
        self,
        controller_type: Type[Controller],
        controller_resource: aconfig.Config,
        result_pipe: Connection,
        *args,
        **kwargs,
    ):
        """Initalize the ReconcileProcessEntrypoint DeployManger and gather start-up configurations

        Args:
            controller_type: Type[Controller]
                The Controller being reconciled
            controller_resource: aconfig.Config
                The resource being reconciled
            result_pipe: Connection
                The pipe to send dependent watch requests to
            *args:
                Extendable arguments to pass to to parent
            **kwargs:
                Extendable key word arguments to pass to parent
        """
        # Initialize ReconcileProcessEntrypoint Deploy Manager
        super().__init__(*args, owner_cr=controller_resource, **kwargs)

        # Initialize required variables
        self.requested_watches = set()
        self.result_pipe = result_pipe
        self.controller_type = controller_type

        # Setup Subsystems
        self.subsystems = self._gather_subsystems(controller_type)
        self.reconcile_manager = ReconcileManager(deploy_manager=self)

    # Functional Overrides

    def _apply_resource(self, resource_definition: dict) -> dict:
        """Override apply resource for handling watch_dependent_resources and subsystem rollout"""
        resource = super()._apply_resource(resource_definition)
        resource_id = ResourceId.from_resource(resource)

        # Send watch request if watch_dependent_resources is enabled
        # and/or handle subsystem rollout
        if config.python_watch_manager.watch_dependent_resources:
            log.debug2("Handling dependent resource %s", resource_id)
            self._handle_dependent_resource(resource_id)

        if (
            config.python_watch_manager.subsystem_rollout
            and resource_id.global_id in self.subsystems
        ):
            log.debug2("Rolling out subsystem %s", resource_id.global_id)
            self._handle_subsystem(
                resource, self.subsystems[resource_id.global_id], False
            )

        return resource

    def _replace_resource(self, resource_definition: dict) -> dict:
        """Override replace resource for handling watch_dependent_resources and subsystem rollout"""
        resource = super()._replace_resource(resource_definition)
        resource_id = ResourceId.from_resource(resource)

        # Send watch request if watch_dependent_resources is enabled
        # and/or handle subsystem rollout
        if config.python_watch_manager.watch_dependent_resources:
            log.debug2("Handling dependent resource %s", resource_id)
            self._handle_dependent_resource(resource_id)

        if (
            config.python_watch_manager.subsystem_rollout
            and resource_id.global_id in self.subsystems
        ):
            log.debug2("Rolling out subsystem %s", resource_id.global_id)
            self._handle_subsystem(
                resource, self.subsystems[resource_id.global_id], False
            )

        return resource

    def _disable(self, resource_definition: dict) -> bool:
        """Override disable to insert subsystem logic"""

        changed = super()._disable(resource_definition)
        if not changed:
            return changed

        resource_id = ResourceId.from_resource(resource_definition)

        # If deleted resource is a subsystem then run reconcile with finalizer
        if (
            config.python_watch_manager.subsystem_rollout
            and resource_id.global_id in self.subsystems
        ):
            success, current_state = self.get_object_current_state(
                kind=resource_id.kind,
                name=resource_id.name,
                namespace=resource_id.namespace,
                api_version=resource_id.api_version,
            )
            if not success or not current_state:
                log.warning(
                    "Unable to fetch owner resource %s/%s/%s/%s",
                    resource_id.kind,
                    resource_id.api_version,
                    resource_id.namespace,
                    resource_id.name,
                )
                return changed

            self._handle_subsystem(
                current_state, self.subsystems[resource_id.global_id], True
            )

        return changed

    def _handle_subsystem(self, resource, controller_type, is_finalizer):
        """Handle rolling out a subsystem for a specific controller, resource, and finalizer"""

        # Copy a ref of the current logging format to restore to
        log_formatters = {}
        for handler in logging.getLogger().handlers:
            log_formatters[handler] = handler.formatter

        # Update the current owner
        current_owner = self._owner_cr
        self._owner_cr = resource
        current_controller_type = self.controller_type
        self.controller_type = controller_type

        # Add the new controllers subsystems to the current dictionary
        # this simplifies future look ups
        current_subsystems = self.subsystems
        self.subsystems = (self._gather_subsystems(controller_type),)

        self.reconcile_manager.safe_reconcile(controller_type, resource, is_finalizer)

        # Reset owner_cr, logging, and subsystems
        self._owner_cr = current_owner
        self.controller_type = current_controller_type
        self.subsystems = current_subsystems
        for handler, formatter in log_formatters.items():
            handler.setFormatter(formatter)

    def _handle_dependent_resource(self, watched_id):
        """Handling request a watch for a deployed resource"""
        # Create requester id
        resource_id = ResourceId.from_resource(self._owner_cr)

        # Remove name from watched_id so it captures
        # any resource of that kind with this owner
        watched_id = copy.deepcopy(watched_id)
        watched_id = dataclasses.replace(watched_id, name=None)

        filters = DependentWatchFilter
        if controller_filters := get_filters_for_resource_id(
            self.controller_type, watched_id
        ):
            filters = AndFilter(DependentWatchFilter, controller_filters)
        watch_filters = FilterManager.to_info(filters)

        watch_request = WatchRequest(
            requester=resource_id,
            watched=watched_id,
            # Use controller info to avoid issues between vcs and pickling
            controller_info=ClassInfo.from_type(self.controller_type),
            filters_info=watch_filters,
        )

        # Only send each watch request once
        if watch_request not in self.requested_watches:
            log.debug3(f"Sending watch request {watch_request}")
            self.result_pipe.send(watch_request)
            self.requested_watches.add(watch_request)

    def _gather_subsystems(self, controller_type: Type[Controller]):
        """Gather the list of subsystems for a controller"""
        subsystem_controllers = getattr(controller_type, "pwm_subsystems", [])
        subsystems = {
            ResourceId.from_controller(controller).global_id: controller
            for controller in subsystem_controllers
        }
        log.debug3(f"Gathered subsystems: {subsystems}")
        return subsystems


## EntryPoint
class ReconcileProcessEntrypoint:  # pylint: disable=too-few-public-methods
    """The ReconcileProcessEntrypoint Class is the main start place for a
    reconciliation. It configures some watch manager specific settings like
    multiprocess logging, and signal handling then it hands off control to the
    ReconcileManager"""

    def __init__(
        self,
        controller_type: Type[Controller],
        deploy_manager: DeployManagerBase = None,
    ):
        """Initializer for the entrypoint class

        Args:
            controller_type: Type[Controller]
                The Controller type being reconciled
            deploy_manager: DeployManagerBase = None
                An optional deploy manager override
        """
        self.controller_type = controller_type
        self.deploy_manager = deploy_manager

        # Initialize the reconcile manager in start
        self.reconcile_manager = None

    def start(
        self,
        logging_queue: multiprocessing.Queue,
        request: ReconcileRequest,
        result_pipe: Connection,
    ):
        """Main entrypoint for the class

        Args:
            logging_queue: multiprocessing.Queue
                The queue to pass log messages to
            request: ReconcileRequest
                The reconcile request that trigger this reconciliation
            result_pipe: Connection
                The connection to send results back to
        """

        # Parse the request and setup local variables
        log.debug4("Setting up resource")
        resource = request.resource
        resource_id = ResourceId.from_resource(resource)

        # Set a unique thread name for each reconcile
        thread_name = f"entrypoint_{resource_id.get_id()}/{resource_id.name}"
        log.debug4("Setting thread name: %s", thread_name)
        threading.current_thread().name = thread_name

        # Reset signal handlers to default function
        log.debug4("Resetting signals")
        signal.signal(signal.SIGINT, signal.SIG_DFL)
        signal.signal(signal.SIGTERM, signal.SIG_DFL)

        # Set the logging library to utilize the multiprocessing logging queue
        log.debug4("Getting root logger")
        root_logger = logging.getLogger()
        log.debug4(
            "Resetting logging handlers (previously: %s)", len(root_logger.handlers)
        )
        root_logger.handlers.clear()

        log.debug4("Creating LogQueueHandler")
        handler = LogQueueHandler(logging_queue, resource)
        log.debug4("Adding handler")
        root_logger.addHandler(handler)

        # Replace stdout and stderr with a null stream as all messages should be passed via
        # the queue and any data in the buffer could cause the process to hang. This can
        # make it difficult to debug subprocesses if they fail before setting up the handler
        log.debug4("Redirecting to /dev/null")
        with open(os.devnull, "w", encoding="utf-8") as null_file:
            sys.stdout = null_file
            sys.stderr = null_file

            log.info(
                "ReconcileProcessEntrypoint for %s and with type: %s",
                self.controller_type,
                request.type,
            )

            # If controller_type has subsystems than set reconciliation to standalone mode.
            # This forces the reconcile to be single threaded but allows for recursive reconciles
            log.debug4("Checking for subsystem rollout")
            if (
                getattr(self.controller_type, "pwm_subsystems", [])
                and config.python_watch_manager.subsystem_rollout
            ):
                config.standalone = True

            # Create a custom deploy manager so we can insert functionality
            deploy_manager = self.deploy_manager
            if not deploy_manager:
                deploy_manager = ReconcileProcessDeployManager(
                    result_pipe=result_pipe,
                    controller_resource=resource.definition,
                    controller_type=self.controller_type,
                )

            # Create a reconciliation manager and start the reconcile
            self.reconcile_manager = ReconcileManager(deploy_manager=deploy_manager)

            finalize = request.type == KubeEventType.DELETED or resource.metadata.get(
                "deletionTimestamp"
            )
            reconcile_result = self.reconcile_manager.safe_reconcile(
                self.controller_type,
                resource.definition,
                finalize,
            )
            # Clear exception as it's not always pickleable
            reconcile_result.exception = None

            # Complete the reconcile by sending the result back up the pipe
            # and explicitly close all remaining descriptors
            log.info("Finished Reconcile for %s", resource_id)
            log.debug3("Sending reconciliation result back to main process")
            result_pipe.send(reconcile_result)
            result_pipe.close()
            logging_queue.close()


def create_and_start_entrypoint(
    logging_queue: multiprocessing.Queue,
    request: ReconcileRequest,
    result_pipe: Connection,
    deploy_manager: DeployManagerBase = None,
):
    """Function to create and start an entrypoint while catching any unexpected errors
    Args:
        logging_queue: multiprocessing.Queue
            The queue to send log messages to
        request: ReconcileRequest
            The request that triggered this reconciliation
        result_pipe: Connection
            The pipe to send a result back with
        deploy_manager: DeployManagerBase = None
            An optional DeployManager override
    """
    try:
        log.debug3("Creating entrypoint")
        entry = ReconcileProcessEntrypoint(
            request.controller_type, deploy_manager=deploy_manager
        )
        log.debug3("Starting entrypoint")
        entry.start(logging_queue, request, result_pipe)
    except Exception as exc:  # pylint: disable=broad-exception-caught
        log.error("Uncaught exception '%s'", exc, exc_info=True)



"""
Test the custom assert functions
"""

# Third Party
import pytest

# Local
from oper8 import exceptions


def test_assert_precondition_pass():
    """Make sure that no exception is throw by assert_precondition when it
    passes
    """
    exceptions.assert_precondition(True)


def test_assert_precondition_fail():
    """Make sure the right exception is thrown by assert_precondition when it
    fails
    """
    exception_msg = "error mesage"
    with pytest.raises(exceptions.PreconditionError, match=exception_msg):
        exceptions.assert_precondition(False, exception_msg)


def test_assert_config_pass():
    """Make sure that no exception is throw by assert_config when it
    passes
    """
    exceptions.assert_config(True)


def test_assert_config_fail():
    """Make sure the right exception is thrown by assert_config when it
    fails
    """
    exception_msg = "error mesage"
    with pytest.raises(exceptions.ConfigError, match=exception_msg):
        exceptions.assert_config(False, exception_msg)


def test_assert_cluster_pass():
    """Make sure that no exception is throw by assert_cluster when it
    passes
    """
    exceptions.assert_cluster(True)


def test_assert_cluster_fail():
    """Make sure the right exception is thrown by assert_cluster when it
    fails
    """
    exception_msg = "error mesage"
    with pytest.raises(exceptions.ClusterError, match=exception_msg):
        exceptions.assert_cluster(False, exception_msg)


def test_exception_derived_from_base():
    """Make sure the derived classes is instance of the base Oper8 Class"""
    fatal_error = exceptions.Oper8FatalError()
    assert isinstance(fatal_error, exceptions.Oper8Error)
    expected_error = exceptions.Oper8ExpectedError()
    assert isinstance(expected_error, exceptions.Oper8Error)


def test_cluter_is_fatal():
    """Make sure the cluster error is considered fatal error"""
    with pytest.raises(exceptions.ClusterError) as cluster_error:
        exceptions.assert_cluster(False)
    assert isinstance(cluster_error.value, exceptions.Oper8FatalError)
    assert cluster_error.value.is_fatal_error


def test_config_is_fatal():
    """Make sure the config error is considered fatal error"""
    with pytest.raises(exceptions.ConfigError) as config_error:
        exceptions.assert_config(False)
    assert isinstance(config_error.value, exceptions.Oper8FatalError)
    assert config_error.value.is_fatal_error


def test_rollout_is_fatal():
    """Make sure the rollout error is considered fatal error"""
    with pytest.raises(exceptions.RolloutError) as rollout_error:
        raise exceptions.RolloutError()
    assert isinstance(rollout_error.value, exceptions.Oper8FatalError)
    assert rollout_error.value.is_fatal_error


def test_precondition_is_non_fatal():
    """Make sure the expected errors are not setting the fatal error flag"""
    with pytest.raises(exceptions.PreconditionError) as precondition_error:
        exceptions.assert_precondition(False)
    assert not precondition_error.value.is_fatal_error
    assert not isinstance(precondition_error.value, exceptions.Oper8FatalError)
    assert isinstance(precondition_error.value, exceptions.Oper8ExpectedError)


def test_verification_is_non_fatal():
    """Make sure the expected errors are not setting the fatal error flag"""
    msg = "Some verification message"
    with pytest.raises(exceptions.VerificationError, match=msg) as verification_error:
        exceptions.assert_verified(False, msg)
    assert not verification_error.value.is_fatal_error
    assert not isinstance(verification_error.value, exceptions.Oper8FatalError)
    assert isinstance(verification_error.value, exceptions.Oper8ExpectedError)



"""
Tests for all functionality of the Session object
"""

# Third Party
import pytest

# First Party
import aconfig

# Local
from oper8.patch import STRATEGIC_MERGE_PATCH
from oper8.session import MAX_NAME_LEN, Session
from oper8.status import make_application_status
from oper8.test_helpers.helpers import (
    DummyNodeComponent,
    MockDeployManager,
    make_patch,
    setup_cr,
)

## Helpers #####################################################################


def make_component_class(comp_name):
    class DerivedComponent(DummyNodeComponent):
        name = comp_name

    return DerivedComponent


def make_api_obj(
    kind="Foo",
    api_version="foo.bar/v1",
    name="foo",
    namespace="testit",
):
    return aconfig.Config(
        {
            "kind": kind,
            "apiVersion": api_version,
            "metadata": {
                "name": name,
                "namespace": namespace,
                "labels": {"app": "test", "run": name},
            },
        },
        override_env_vars=False,
    )


## Tests #######################################################################

###############
## Prperties ##
###############


def test_constructed_properties():
    """Make sure all properties derived from the constructor args are populated
    correctly
    """
    rec_id = "1ab"
    cr = setup_cr()
    cfg = aconfig.Config({"foo": "bar"}, override_env_vars=False)
    dm = MockDeployManager()
    patches = [make_patch(STRATEGIC_MERGE_PATCH, {})]
    session = Session(rec_id, cr, cfg, dm, patches)
    assert session.id == rec_id
    assert session.cr_manifest == cr
    assert session.config == cfg
    assert session.deploy_manager == dm
    assert session.temporary_patches == patches


def test_cr_properties():
    """Make sure all properties derived from the CR manifest are populated
    correctly
    """
    version = "develop.1.2.3"
    namespace = "wingbat"
    name = "wombat"
    api_version = "critters.bats/v23"
    kind = "Critter"
    spec = {"key": "value"}
    cr = setup_cr(
        api_version=api_version,
        kind=kind,
        namespace=namespace,
        name=name,
        version=version,
        spec=spec,
    )
    session = Session("1ab", cr, {}, MockDeployManager())
    assert session.version == version
    assert session.namespace == namespace
    assert session.name == name
    assert session.kind == kind
    assert session.api_version == api_version
    assert session.spec == cr.spec
    assert session.metadata == cr.metadata


@pytest.mark.parametrize(
    "field",
    ["kind", "apiVersion", "metadata", "metadata.name", "metadata.namespace"],
)
def test_missing_cr_required_fields(field):
    """Make sure that required fields missing from the CR correctly raise
    validation errors
    """
    cr = setup_cr()
    field_parts = field.split(".")
    dct = cr
    while len(field_parts) > 1:
        dct = dct[field_parts[0]]
        field_parts = field_parts[1:]
    del dct[field_parts[0]]
    with pytest.raises(AssertionError):
        session = Session("1ab", cr, {}, MockDeployManager())


@pytest.mark.parametrize(
    "field,expected",
    [("spec", aconfig.Config({})), ("spec.version", None)],
)
def test_missing_cr_optional_fields(field, expected):
    """Make sure that optional fields in the CR that are accessed via properties
    do not raise errors
    """
    cr = setup_cr()
    field_parts = field.split(".")
    dct = cr
    while len(field_parts) > 1:
        dct = dct[field_parts[0]]
        field_parts = field_parts[1:]
    del dct[field_parts[0]]
    session = Session("1ab", cr, {}, MockDeployManager())
    assert getattr(session, field.split(".")[-1]) == expected


def test_current_version():
    """Make sure that retrieving the current_version works when it's present
    in the deploy manager
    """
    # Make sure that current_version is not set when it hasn't been deployed
    cr = setup_cr()
    dm = MockDeployManager()
    session = Session("1ab", cr, {}, dm)
    assert session.current_version is None

    # Make sure that current_version is set when it's been deployed before
    current_version = "some-version"
    cr.status = make_application_status(version=current_version)
    dm.deploy([cr])
    session = Session("2cd", cr, {}, dm)
    assert session.current_version == current_version


def test_dict_conversion():
    """Make sure that a CR as a raw dict is converted (for that 100% cov!)"""
    Session("1ab", dict(setup_cr()), {}, MockDeployManager())


######################
## State Management ##
######################


def test_add_component_ok():
    """Make sure that add_component successfully adds enabled and disabled
    components
    """
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    enabled_comp = make_component_class("enabled")(session)
    disabled_comp = make_component_class("disabled")(session, disabled=True)
    assert session.get_components() == [enabled_comp]
    assert session.get_components(disabled=True) == [disabled_comp]


def test_add_component_no_duplicate():
    """Make sure that the same component cannot be added multiple times"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    CompType = make_component_class("comp")
    CompType(session)
    with pytest.raises(ValueError):
        CompType(session)


def test_add_component_dependency_ok():
    """Make sure that adding a dependency from one component to another works as
    expected
    """
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    comp1 = make_component_class("one")(session)
    comp2 = make_component_class("two")(session)
    session.add_component_dependency(comp1, comp2)
    assert session.get_component_dependencies(comp1.name) == [(comp2, None)]


def test_add_component_dependency_verify_function():
    """Make sure that adding a dependency with a verify function stores the
    verify function successfully
    """
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    comp1 = make_component_class("one")(session)
    comp2 = make_component_class("two")(session)
    verify_fn = lambda *_, **__: True
    session.add_component_dependency(comp1, comp2, verify_function=verify_fn)
    dep_list = session.get_component_dependencies(comp1.name)
    assert len(dep_list) == 1
    assert dep_list[0][0] == comp2
    assert dep_list[0][1].func == verify_fn


def test_add_component_dependency_by_name():
    """Make sure that adding a dependency can be done by name"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    comp1 = make_component_class("one")(session)
    comp2 = make_component_class("two")(session)
    session.add_component_dependency(comp1.name, comp2.name)
    assert session.get_component_dependencies(comp1.name) == [(comp2, None)]


def test_add_component_dependency_no_unknown_component():
    """Make sure that a dependency can't be added for an unknown component"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    comp2 = make_component_class("two")(session)
    with pytest.raises(ValueError):
        session.add_component_dependency("one", comp2)


def test_add_component_dependency_no_unknown_upstream():
    """Make sure that a dependency can't be added for an unknown upstream"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    comp1 = make_component_class("one")(session)
    with pytest.raises(ValueError):
        session.add_component_dependency(comp1, "two")


def test_add_component_dependency_no_disabled_component():
    """Make sure that a dependency can't be added for a disabled component"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    enabled_comp = make_component_class("enabled")(session)
    disabled_comp = make_component_class("disabled")(session, disabled=True)
    with pytest.raises(ValueError):
        session.add_component_dependency(disabled_comp, enabled_comp)


def test_add_component_dependency_no_disabled_upstream():
    """Make sure that a dependency can't be added for a disabled upstream"""
    session = Session("1ab", setup_cr(), {}, MockDeployManager())
    enabled_comp = make_component_class("enabled")(session)
    disabled_comp = make_component_class("disabled")(session, disabled=True)
    with pytest.raises(ValueError):
        session.add_component_dependency(enabled_comp, disabled_comp)


###############
## Utilities ##
###############


def test_get_scoped_name():
    """Make sure that get_scoped_name correctly scopes and truncates"""
    name = "some-name"
    dm = MockDeployManager()
    session = Session("1ab", setup_cr(name=name), {}, dm)

    # Test a name shorter than the max length
    assert session.get_scoped_name("foo") == f"{name}-foo"

    # Test a name longer than the max length
    scoped_name = session.get_scoped_name(
        "".join((str(i % 10) for i in range(MAX_NAME_LEN)))
    )
    assert len(scoped_name) == MAX_NAME_LEN
    assert scoped_name.startswith(name)


def test_get_truncated_name():
    """Make sure that get_truncated_name correctly (only) truncates"""
    name = "some-name"
    dm = MockDeployManager()
    session = Session("1ab", setup_cr(name=name), {}, dm)

    # Test a name shorter than the max length
    assert session.get_truncated_name("foo") == "foo"

    # Test a name longer than the max length
    long_name = "".join((str(i % 10) for i in range(MAX_NAME_LEN * 2)))
    trunc_name = session.get_truncated_name(long_name)
    assert len(trunc_name) == MAX_NAME_LEN
    assert not trunc_name.startswith(name)

    # Test that truncation makes names unique when the instance name is longer
    # than the max
    session = Session("1ab", setup_cr(name=long_name), {}, dm)
    assert session.get_truncated_name("foo") != session.get_truncated_name("bar")


def test_get_object_current_state():
    """Make sure that get_object_current_state correctly proxies to the deploy
    manager with the current namespace
    """
    dm = MockDeployManager()
    session = Session("1ab", setup_cr(), {}, dm)
    obj1 = make_api_obj(namespace=session.namespace)
    obj2 = make_api_obj(namespace="different-namespace", name="other-name")
    dm.deploy([obj1, obj2])

    # Make sure it fetches correctly with api_version in the same namespace
    success, obj1_state = session.get_object_current_state(
        kind=obj1.kind,
        api_version=obj1.apiVersion,
        name=obj1.metadata.name,
    )
    assert success
    assert obj1_state == obj1

    # Make sure it fetches correctly without api_version in the same namespace
    success, obj1_state = session.get_object_current_state(
        kind=obj1.kind,
        name=obj1.metadata.name,
    )
    assert success
    assert obj1_state == obj1

    # Make sure it doesn't find objects in different namespaces
    success, obj2_state = session.get_object_current_state(
        kind=obj2.kind,
        api_version=obj2.apiVersion,
        name=obj2.metadata.name,
    )
    assert success
    assert obj2_state is None


def test_filter_objects_current_state():
    """Make sure that filter_object_current_state correctly proxies to the deploy
    manager with the current namespace
    """
    dm = MockDeployManager()
    session = Session("1ab", setup_cr(), {}, dm)
    obj1 = make_api_obj(namespace=session.namespace, name="first")
    obj2 = make_api_obj(namespace=session.namespace, name="second")
    obj3 = make_api_obj(namespace="different-namespace", name="other-name")
    dm.deploy([obj1, obj2, obj3])

    # Make sure it fetches correctly in the same namespace
    success, obj_list = session.filter_objects_current_state(
        kind=obj1.kind,
        api_version=obj1.apiVersion,
    )
    assert success
    assert len(obj_list) == 2
    assert obj1 in obj_list and obj2 in obj_list

    # Make sure it fetches correctly filters labels
    success, obj_list = session.filter_objects_current_state(
        kind=obj1.kind, label_selector="run=first"
    )
    assert success
    assert len(obj_list) == 1
    assert obj_list[0] == obj1

    # Make sure it fetches correctly filters fields
    success, obj_list = session.filter_objects_current_state(
        kind=obj1.kind, field_selector="metadata.name!=first"
    )
    assert success
    assert len(obj_list) == 1
    assert obj_list[0] == obj2

    # Make sure it doesn't find objects in different namespaces
    success, obj_list = session.filter_objects_current_state(
        kind=obj2.kind, field_selector="metadata.name=other-name"
    )
    assert success
    assert len(obj_list) == 0



"""
Test the verify_resources functionality
"""

# Standard
from datetime import datetime, timedelta
from typing import Optional

# First Party
import alog

# Local
from oper8 import status
from oper8.session import _SESSION_NAMESPACE
from oper8.test_helpers.helpers import (
    MockDeployManager,
    configure_logging,
    setup_session,
)
from oper8.utils import nested_set
from oper8.verify_resources import (
    AVAILABLE_CONDITION_KEY,
    COMPLETE_CONDITION_KEY,
    DEFAULT_TIMESTAMP_KEY,
    NEW_RS_AVAILABLE_REASON,
    PROGRESSING_CONDITION_KEY,
    verify_resource,
)

configure_logging()
log = alog.use_channel("TEST")

## Helpers #####################################################################

TEST_NAMESPACE = "test"


def run_test_verify(
    kind,
    conditions=None,
    status=None,
    populate_state=True,
    version="1.2.3",
    obj_namespace=TEST_NAMESPACE,
    search_namespace=TEST_NAMESPACE,
    **kwargs
):
    """Common helper for all tests"""

    # We're not testing the lookup functionality, so just use a canned name and
    # api_version
    name = "foo"
    api_version = "v1"

    # Set up the state of the cluster
    object_state = {}
    if populate_state:
        object_state = {
            "apiVersion": api_version,
            "kind": kind,
            "metadata": {
                "name": name,
                "namespace": obj_namespace,
            },
        }
        if conditions is not None:
            status = status or {}
            status.update({"conditions": conditions})

        if status is not None:
            object_state["status"] = status

    # Set up a mock deploy manager with the seed state
    dm = MockDeployManager(resources=[object_state])

    # Set up the session with the state
    session = setup_session(
        namespace=TEST_NAMESPACE, deploy_manager=dm, version=version
    )

    # Run the verify call on the resource
    return verify_resource(
        kind=kind,
        name=name,
        api_version=api_version,
        session=session,
        namespace=search_namespace,
        **kwargs
    )


def make_condition(
    type_name,
    status,
    timestamp=None,
    timestamp_key=None,
    timestamp_str=True,
    reason: Optional[str] = None,
):
    """Helper for making conditions easily"""
    timestamp_key = timestamp_key or DEFAULT_TIMESTAMP_KEY
    if timestamp_str:
        timestamp = timestamp or datetime.now().isoformat()
    return {
        "type": type_name,
        "status": str(status),
        timestamp_key: timestamp,
        "reason": reason,
    }


def make_status_with_version():
    out = {}
    nested_set(out, status.VERSIONS_FIELD_CURRENT_VERSION, "1.0.0")
    return out


## Tests #######################################################################

##########
## Pods ##
##########


def test_verify_pod_ready():
    """Make sure a ready pod verifies cleanly"""
    assert run_test_verify(kind="Pod", conditions=[make_condition("Ready", True)])


def test_verify_pod_unready():
    """Make sure an unready pod returns as unverified"""
    assert not run_test_verify(kind="Pod", conditions=[make_condition("Ready", False)])


def test_verify_pod_missing():
    """Make sure a missing pod returns as unverified"""
    assert not run_test_verify(kind="Pod", populate_state=None)


def test_verify_pod_no_conditions():
    """Make sure a pod with no conditions returns as unverified"""
    assert not run_test_verify(kind="Pod", conditions=[])


def test_verify_pod_no_status():
    """Make sure a pod with no status returns as unverified"""
    assert not run_test_verify(kind="Pod", conditions=None)


def test_verify_pod_separate_namespace():
    """Make sure a ready pod from a different namespace verifies cleanly"""
    assert run_test_verify(
        kind="Pod",
        conditions=[make_condition("Ready", True)],
        obj_namespace="adifferent",
        search_namespace="adifferent",
    )


def test_verify_pod_null_namespace():
    """Make sure a ready pod in the same namespace verifies cleanly"""
    assert run_test_verify(
        kind="Pod",
        conditions=[make_condition("Ready", True)],
        obj_namespace=TEST_NAMESPACE,
        search_namespace=_SESSION_NAMESPACE,
    )


def test_verify_pod_custom_verification():
    """Make sure a ready pod fails to verify with a custom override"""
    assert not run_test_verify(
        kind="Pod",
        conditions=[make_condition("Ready", True)],
        verify_function=lambda resource: False,
    )


##########
## Jobs ##
##########


def test_verify_job_completed():
    """Make sure a completed job verifies cleanly"""
    assert run_test_verify(
        kind="Job", conditions=[make_condition(COMPLETE_CONDITION_KEY, True)]
    )


def test_verify_job_failed():
    """Make sure a failed job returns as unverified"""
    assert not run_test_verify(kind="Job", conditions=[make_condition("Failed", True)])


def test_verify_job_suspended():
    """Make sure a suspended job returns as unverified"""
    assert not run_test_verify(
        kind="Job", conditions=[make_condition("Suspended", True)]
    )


def test_verify_job_missing():
    """Make sure a missing job returns as unverified"""
    assert not run_test_verify(kind="Job", populate_state=None)


def test_verify_job_no_conditions():
    """Make sure a job with no conditions returns as unverified"""
    assert not run_test_verify(kind="Job", conditions=[])


def test_verify_job_no_status():
    """Make sure a job with no status returns as unverified"""
    assert not run_test_verify(kind="Job", conditions=None)


def test_verify_job_separate_namespace():
    """Make sure a completed job from a different namespace verifies cleanly"""
    assert run_test_verify(
        kind="Job",
        conditions=[make_condition(COMPLETE_CONDITION_KEY, True)],
        obj_namespace="adifferent",
        search_namespace="adifferent",
    )


def test_verify_job_null_namespace():
    """Make sure a completed job in the same namespace verifies cleanly"""
    assert run_test_verify(
        kind="Job",
        conditions=[make_condition(COMPLETE_CONDITION_KEY, True)],
        obj_namespace=TEST_NAMESPACE,
        search_namespace=_SESSION_NAMESPACE,
    )


#################
## Deployments ##
#################


def test_verify_deploy_stable():
    """Make sure a deploy with all of the expected replicas verifies cleanly"""
    assert run_test_verify(
        kind="Deployment",
        conditions=[
            make_condition(AVAILABLE_CONDITION_KEY, True),
            make_condition(
                PROGRESSING_CONDITION_KEY, True, reason=NEW_RS_AVAILABLE_REASON
            ),
        ],
    )


def test_verify_deploy_unavailable():
    """Make sure a deploy without all of the expected replicas returns as
    unverified
    """
    assert not run_test_verify(
        kind="Deployment", conditions=[make_condition(AVAILABLE_CONDITION_KEY, False)]
    )


def test_verify_deploy_rollout_uncompleted():
    """Make sure a deploy is not verified while rolling out to new version
    is still in-progress
    """
    assert not run_test_verify(
        kind="Deployment",
        conditions=[
            make_condition(AVAILABLE_CONDITION_KEY, True),
            make_condition(PROGRESSING_CONDITION_KEY, True, reason="ReplicaSetUpdated"),
        ],
    )


def test_verify_deploy_missing():
    """Make sure a missing deploy returns as unverified"""
    assert not run_test_verify(kind="Deployment", populate_state=None)


def test_verify_deploy_no_conditions():
    """Make sure a deploy with no conditions returns as unverified"""
    assert not run_test_verify(kind="Deployment", conditions=[])


def test_verify_deploy_no_status():
    """Make sure a deploy with no status returns as unverified"""
    assert not run_test_verify(kind="Deployment", conditions=None)


##################
## StatefulSets ##
##################


def test_verify_ss_available():
    """Make sure a statefulset with all of the expected replicas verifies
    cleanly
    """
    assert run_test_verify(
        kind="StatefulSet", status={"replicas": 3, "readyReplicas": 3}
    )


def test_verify_ss_some_ready():
    """Make sure a statefulset with only some replicas ready is unverified"""
    assert not run_test_verify(
        kind="StatefulSet", status={"replicas": 3, "readyReplicas": 2}
    )


def test_verify_ss_none_ready():
    """Make sure a statefulset with no readyReplicas is unverified"""
    assert not run_test_verify(kind="StatefulSet", status={"replicas": 3})


def test_verify_ss_missing():
    """Make sure a missing deploy returns as unverified"""
    assert not run_test_verify(kind="StatefulSet", populate_state=None)


def test_verify_ss_no_status():
    """Make sure a deploy with no conditions returns as unverified"""
    assert not run_test_verify(kind="StatefulSet")


################
## Subsystems ##
################


def test_verify_subsystem_ready():
    """Make sure a ready subsystem verifies cleanly"""
    assert run_test_verify(
        kind="FooBar",
        is_subsystem=True,
        version="1.0.0",
        status=make_status_with_version(),
        conditions=[
            make_condition(
                status.READY_CONDITION, True, timestamp_key=status.TIMESTAMP_KEY
            ),
            make_condition(
                status.UPDATING_CONDITION, False, timestamp_key=status.TIMESTAMP_KEY
            ),
        ],
    )


def test_verify_subsystem_unready():
    """Make sure an unready subsystem verifies returns as unverified"""
    assert not run_test_verify(
        kind="FooBar",
        is_subsystem=True,
        version="1.0.0",
        status=make_status_with_version(),
        conditions=[
            make_condition(
                status.READY_CONDITION, False, timestamp_key=status.TIMESTAMP_KEY
            )
        ],
    )


def test_verify_subsystem_unready_updating_in_progress():
    """Make sure a subsystem is not verified during update"""
    assert not run_test_verify(
        kind="FooBar",
        is_subsystem=True,
        version="1.0.0",
        status=make_status_with_version(),
        conditions=[
            make_condition(
                status.READY_CONDITION, True, timestamp_key=status.TIMESTAMP_KEY
            ),
            make_condition(
                status.UPDATING_CONDITION, True, timestamp_key=status.TIMESTAMP_KEY
            ),
        ],
    )


def test_verify_subsystem_unready_version_mismatch():
    """Make sure a subsystem is not verified in case updating is not started
    by checking current version and desired version difference.
    """
    assert not run_test_verify(
        kind="FooBar",
        is_subsystem=True,
        version="2.0.0",
        status=make_status_with_version(),
        conditions=[
            make_condition(
                status.READY_CONDITION, True, timestamp_key=status.TIMESTAMP_KEY
            ),
            make_condition(
                status.UPDATING_CONDITION, False, timestamp_key=status.TIMESTAMP_KEY
            ),
        ],
    )


def test_verify_subsystem_unready_no_version():
    """Make sure a subsystem is not verified in case version status is not set
    as initial rollout is not completed.
    """
    assert not run_test_verify(
        kind="FooBar",
        is_subsystem=True,
        version="2.0.0",
        conditions=[
            make_condition(
                status.READY_CONDITION, True, timestamp_key=status.TIMESTAMP_KEY
            ),
            make_condition(
                status.UPDATING_CONDITION, False, timestamp_key=status.TIMESTAMP_KEY
            ),
        ],
    )


def test_verify_subsystem_missing():
    """Make sure a missing subsystem returns as unverified"""
    assert not run_test_verify(kind="FooBar", is_subsystem=True, populate_state=False)


def test_verify_subsystem_no_conditions():
    """Make sure a subsystem with no conditions returns as unverified"""
    assert not run_test_verify(kind="FooBar", is_subsystem=True, conditions=[])


def test_verify_subsystem_no_status():
    """Make sure a subsystem with no status returns as unverified"""
    assert not run_test_verify(kind="FooBar", is_subsystem=True, conditions=None)


#######################
## Custom Conditions ##
#######################


def test_custom_condition_true():
    """Make sure that verification with a custom condition that is true returns
    as verified
    """
    assert run_test_verify(
        kind="WingBat",
        condition_type="Custom",
        conditions=[make_condition("Custom", True)],
    )


def test_custom_condition_true_custom_timestamp():
    """Make sure that verification with a custom condition and a custom
    timestamp key that is true returns as verified
    """
    assert run_test_verify(
        kind="WingBat",
        condition_type="Custom",
        timestamp_key="custom",
        conditions=[make_condition("Custom", True, timestamp_key="custom")],
    )


def test_custom_condition_false():
    """Make sure that verification with a custom condition that is false returns
    as unverified
    """
    assert not run_test_verify(
        kind="WingBat",
        condition_type="Custom",
        conditions=[make_condition("Custom", False)],
    )


def test_custom_condition_missing():
    """Make sure that verification with a custom condition that is missing
    returns as unverified
    """
    assert not run_test_verify(
        kind="WingBat", condition_type="Custom", populate_state=False
    )


###########
## Other ##
###########


def test_multiple_condition_entries():
    """Make sure that a resource with multiple condition entries for the same
    type uses the latest
    """
    first_ts = datetime.now()
    t0 = first_ts.isoformat()
    t1 = (first_ts + timedelta(minutes=1)).isoformat()
    assert run_test_verify(
        kind="Pod",
        conditions=[
            make_condition("Ready", False, timestamp=t0),
            make_condition("Ready", True, timestamp=t1),
        ],
    )


def test_non_bool_str_value():
    """Make sure that a condition with a string value that doesn't look like a
    bool returns as unverified
    """
    assert not run_test_verify(
        kind="WingBat",
        condition_type="Custom",
        conditions=[make_condition("Custom", "NotABool")],
    )


def test_generic_resource_present():
    """Make sure that a resource without a known verifier returns true when
    present
    """
    assert run_test_verify(kind="Service")


def test_generic_resource_missing():
    """Make sure that a resource without a known verifier returns false when
    not present
    """
    assert not run_test_verify(kind="Service", populate_state=False)


def test_condition_missing_status():
    """Make sure that a resource with a condition that has no status returns as
    unverified
    """
    assert not run_test_verify(kind="Pod", conditions=[{"type": "Ready"}])


def test_condition_non_str_status():
    """Make sure that a non-string status is handled by casting to bool"""
    assert run_test_verify(kind="Pod", conditions=[{"type": "Ready", "status": True}])


def test_datetime_timestamp():
    """Make sure that when a timestamp field has been parsed as a datetime
    object by the yaml parser, it is handled correctly
    """
    assert run_test_verify(
        kind="Pod",
        conditions=[
            make_condition(
                "Ready", True, timestamp=datetime.now(), timestamp_str=False
            ),
        ],
    )



"""
Shared test config
"""
# Standard
from unittest import mock
import sys

# Third Party
import pytest

# Local
from oper8.reconcile import ReconcileManager
from oper8.test_helpers.helpers import configure_logging, version_safe_md5

configure_logging()


@pytest.fixture(autouse=True)
def no_local_kubeconfig():
    """This fixture makes sure the tests run as if KUBECONFIG is not exported in
    the environment, even if it is
    """
    with mock.patch(
        "kubernetes.config.new_client_from_config", side_effect=RuntimeError
    ):
        yield


@pytest.fixture(autouse=True)
def no_unimport_oper8_mods():
    """Since our helper classes themselves live within the top of oper8, the
    real logic for unimporting the controller modules will recursively reimport
    _all_ oper8.* modules. This causes a lot of problems with tests like
    misbehaving mocks, Controller is not Controller, etc...
    """
    real_unimport = ReconcileManager._unimport_controller_module

    @staticmethod
    def _patched_unimport(module_name):
        oper8_mods = {
            mod_name: mod
            for mod_name, mod in sys.modules.items()
            if mod_name.startswith("oper8.")
            and not mod_name.startswith("oper8.test_helpers")
        }
        reimport_modules = real_unimport(module_name)
        for mod_name, mod in oper8_mods.items():
            sys.modules.setdefault(mod_name, mod)
        return reimport_modules

    with mock.patch(
        "oper8.reconcile.ReconcileManager._unimport_controller_module",
        new=_patched_unimport,
    ):
        yield



"""
Helper methods for tls factory tests / TLC context tests.
"""

# Standard
import os
import re

# First Party
import alog

# Local
from oper8 import Session, assert_precondition
from oper8.test_helpers.oper8x_helpers import TEST_DATA_DIR, set_secret_data
from oper8.x.oper8x_component import Oper8xComponent
from oper8.x.utils import common
from oper8.x.utils.tls_context.factory import (
    _TlsContextSingletonFactory,
    get_tls_context,
)

log = alog.use_channel("TEST")

CERT_MGR_TEST_TLS_CERT = "-----BEGIN CERTIFICATE-----\nfake-cert"
CERT_MGR_TEST_TLS_KEY = "-----BEGIN FAKED PRIVATE KEY-----\nfake-key"
CERT_MGR_TEST_CA_CERT = "-----BEGIN CERTIFICATE-----\nfake-cacert"

NAME_CLEANER = re.compile("[^-a-zA-Z0-9]")


def reset_tls_factory():
    _TlsContextSingletonFactory._instance = None


def make_server_component_class(server_name="test-server", request_cert=True):
    class ServerComponent(Oper8xComponent):
        name = server_name

        KEY_FIELD = "copied_server_key"
        CRT_FIELD = "copied_server_crt"

        def __init__(self, session, *args, **kwargs):
            super().__init__(session, *args, **kwargs)

            # Get a server key/cert pair to use
            if request_cert:
                log.debug("[%s] Fetching TLS content", self)
                self.server_key, self.server_crt = get_tls_context(
                    self.session
                ).get_server_key_cert_pair(self, [], encode=True)
                log.debug3("Server Key: %s", self.server_key)
                log.debug3("Server Crt: %s", self.server_crt)

        def build_chart(self, *_, **__):
            log.debug("[%s] build_chart", self)
            assert_precondition(None not in [self.server_key, self.server_crt])

            # Add a secret to indicate that the precondition passed
            secret_name = self.get_secret_name()
            self.add_resource(
                name=secret_name,
                obj=dict(
                    kind="Secret",
                    apiVersion="v1",
                    metadata=dict(name=secret_name),
                    data={
                        self.KEY_FIELD: self.server_key,
                        self.CRT_FIELD: self.server_crt,
                    },
                ),
            )

        @classmethod
        def get_secret_name(cls):
            return f"{cls.name}-secret"

    return ServerComponent


def set_cert_manager_secret_for_component(
    session: Session,
    component_name: str,
    data_override={},
    name_override: str = None,
):
    component_name = NAME_CLEANER.sub("", component_name).lower()
    if not data_override:
        data_override = {
            "tls.key": common.b64_secret(CERT_MGR_TEST_TLS_KEY),
            "tls.crt": common.b64_secret(CERT_MGR_TEST_TLS_CERT),
            "ca.crt": common.b64_secret(CERT_MGR_TEST_CA_CERT),
        }
    set_secret_data(
        session,
        name=("tls-" + component_name) if name_override is None else name_override,
        data=data_override,
        secret_type="kubernetes.io/tls",
    )
    set_cert_manager_secret_ca(session)


def set_cert_manager_secret_ca(session: Session):
    with open(os.path.join(TEST_DATA_DIR, "test_ca.key"), "r") as f:
        key_pem = f.read()
    with open(os.path.join(TEST_DATA_DIR, "test_ca.crt"), "r") as f:
        crt_pem = f.read()
    ca_data = {
        "tls.key": common.b64_secret(key_pem),
        "tls.crt": common.b64_secret(crt_pem),
        "ca.crt": common.b64_secret(crt_pem),
    }
    set_secret_data(session, name="ca", data=ca_data, secret_type="kubernetes.io/tls")



"""
Tests for the TimerThread
"""
# Standard
from datetime import datetime, timedelta
import time

# Third Party
import pytest

# Local
from oper8.test_helpers.pwm_helpers import MockedTimerThread

## Helpers #####################################################################


class Counter:
    def __init__(self, initial_value=0):
        self.value = initial_value

    def increment(self, value=1):
        self.value += value


@pytest.mark.timeout(5)
def test_timer_thread_happy_path():
    timer = MockedTimerThread()
    timer.start_thread()

    value_tracker = Counter()
    timer.put_event(datetime.now(), value_tracker.increment)
    timer.put_event(datetime.now() + timedelta(seconds=0.1), value_tracker.increment)
    timer.put_event(datetime.now() + timedelta(seconds=0.2), value_tracker.increment, 2)
    timer.put_event(
        datetime.now() + timedelta(seconds=0.3), value_tracker.increment, value=2
    )
    time.sleep(2.5)
    timer.stop_thread()
    assert value_tracker.value == 6


@pytest.mark.timeout(5)
def test_timer_thread_canceled():
    timer = MockedTimerThread()

    value_tracker = Counter()
    timer.put_event(datetime.now(), value_tracker.increment)
    canceled_event = timer.put_event(
        datetime.now() + timedelta(seconds=0.5), value_tracker.increment
    )
    canceled_event.cancel()

    timer.start_thread()
    time.sleep(2)
    timer.stop_thread()
    assert value_tracker.value == 1
